import operator
from collections.abc import Mapping
from math import gcd

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _broadcast_to,
    _ensure_array,
    _is_negative_zero,
    _is_positive_zero,
    _is_sign_negative_number,
    _is_sign_positive_number,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
    _stack,
    _where,
)
from ....utils.bindings import Parameter
from ..bound import (
    checked_data_bounds,
    guarantee_arg_within_expr_bounds,
    guarantee_stacked_arg_within_expr_bounds,
)
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant
from .literal import Number


class ScalarMultiply(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    def __new__(cls, a: AnyExpr, b: AnyExpr) -> "ScalarMultiply | Number":  # type: ignore[misc]
        ab = Number.symbolic_fold_binary(a, b, operator.mul)
        if ab is not None:
            return ab
        return super().__new__(cls)

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarMultiply | Number":
        return ScalarMultiply(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a, self._b, dtype, np.multiply, ScalarMultiply
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.multiply(self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a_const = not self._a.has_data
        b_const = not self._b.has_data
        assert not (a_const and b_const), "constant multiplication has no data bounds"

        # evaluate a and b and a*b
        a, b = self._a, self._b
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)
        exprv = np.multiply(av, bv)

        if a_const or b_const:
            term, termv, constv = (b, bv, av) if a_const else (a, av, bv)

            fmax = np.finfo(Xs.dtype).max
            smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

            # for x*0, we can allow any finite x, unless the output +-0.0 sign
            #  is restricted, then we need to restrict the sign of x:
            #   | constv | expr_lower | expr_upper | term_lower | term_upper |
            #   | +0.0   | >= +0.0    | >= +0.0    | +0.0       | +fmax      |
            #   | +0.0   | <= -0.0    | <= -0.0    | -fmax      | -0.0       |
            #   | +0.0   | <= -0.0    | >= +0.0    | -fmax      | +fmax      |
            #   | -0.0   | >= +0.0    | >= +0.0    | -fmax      | -0.0       |
            #   | -0.0   | <= -0.0    | <= -0.0    | +0.0       | +fmax      |
            #   | -0.0   | <= -0.0    | >= +0.0    | -fmax      | +fmax      |
            #  - term_lower := +0.0 if
            #    signbit(constv) == signbit(expr_lower) == signbit(expr_upper)
            #  - term_upper := -0.0 if
            #    ~signbit(constv) == signbit(expr_lower) == signbit(expr_upper)
            # for x*Inf, we can allow any non-zero non-NaN x with the same sign
            # for x*NaN, we can allow any x but only propagate [-inf, inf]
            #  since [-NaN, NaN] would be misunderstood as only NaN
            # if term_lower == termv and termv == -0.0, we need to guarantee
            #  that term_lower is also -0.0, same for term_upper
            term_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
                expr_lower, copy=True
            )
            np.copyto(
                term_lower,
                expr_upper,
                where=_is_sign_negative_number(constv),
                casting="no",
            )
            np.divide(term_lower, constv, out=term_lower)
            term_lower[np.isnan(constv)] = -np.inf
            term_lower[np.isinf(constv)] = smallest_subnormal
            term_lower[np.isinf(constv) & _is_sign_negative_number(termv)] = -np.inf
            np.copyto(
                term_lower, termv, where=(np.isinf(constv) & (termv == 0)), casting="no"
            )
            term_lower[constv == 0] = -fmax
            term_lower[
                (constv == 0)
                & (_is_positive_zero(constv) == _is_sign_positive_number(expr_lower))
                & (_is_positive_zero(constv) == _is_sign_positive_number(expr_upper))
            ] = +0.0
            term_lower = _ensure_array(_minimum_zero_sign_sensitive(termv, term_lower))

            term_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
                expr_upper, copy=True
            )
            np.copyto(
                term_upper,
                expr_lower,
                where=_is_sign_negative_number(constv),
                casting="no",
            )
            np.divide(term_upper, constv, out=term_upper)
            term_upper[np.isnan(constv)] = np.inf
            term_upper[np.isinf(constv)] = np.inf
            term_upper[
                np.isinf(constv) & _is_sign_negative_number(termv)
            ] = -smallest_subnormal
            np.copyto(
                term_upper, termv, where=(np.isinf(constv) & (termv == 0)), casting="no"
            )
            term_upper[constv == 0] = fmax
            term_upper[
                (constv == 0)
                & (_is_negative_zero(constv) != _is_sign_negative_number(expr_lower))
                & (_is_negative_zero(constv) != _is_sign_negative_number(expr_upper))
            ] = -0.0
            term_upper = _ensure_array(_maximum_zero_sign_sensitive(termv, term_upper))

            # we need to force argv if expr_lower == expr_upper and constv is
            #  finite non-zero (in other cases we explicitly expand ranges)
            np.copyto(
                term_lower,
                termv,
                where=(
                    (expr_lower == expr_upper) & np.isfinite(constv) & (constv != 0)
                ),
                casting="no",
            )
            np.copyto(
                term_upper,
                termv,
                where=(
                    (expr_lower == expr_upper) & np.isfinite(constv) & (constv != 0)
                ),
                casting="no",
            )

            # handle rounding errors in multiply(divide(...)) early
            term_lower = guarantee_arg_within_expr_bounds(
                lambda term_lower: np.multiply(term_lower, constv),
                exprv,
                termv,
                term_lower,
                expr_lower,
                expr_upper,
            )
            term_upper = guarantee_arg_within_expr_bounds(
                lambda term_upper: np.multiply(term_upper, constv),
                exprv,
                termv,
                term_upper,
                expr_lower,
                expr_upper,
            )

            return term.compute_data_bounds(
                term_lower,
                term_upper,
                Xs,
                late_bound,
            )

        # if neither a not b is const, we simplify by not allowing the sign of
        #  a*b to change
        # we further simplify the code by working with abs(a*b) = abs(a)*abs(b)
        #  and only taking the sign of a and b into account at the end

        # limit expr_lower and expr_upper to keep the same sign
        # then compute the lower and upper bounds for the abs(expr)
        expr_lower = _ensure_array(expr_lower, copy=True)
        expr_lower[_is_sign_positive_number(exprv) & (expr_lower <= 0)] = +0.0
        expr_upper = _ensure_array(expr_upper, copy=True)
        expr_upper[_is_sign_negative_number(exprv) & (expr_upper >= 0)] = -0.0
        expr_abs_lower, expr_abs_upper = (
            _ensure_array(
                _where(_is_sign_negative_number(exprv), -expr_upper, expr_lower)
            ),
            _ensure_array(
                _where(_is_sign_negative_number(exprv), -expr_lower, expr_upper)
            ),
        )

        av_abs = np.abs(av)
        bv_abs = np.abs(bv)
        exprv_abs = _ensure_array(np.abs(exprv))

        fmax = np.finfo(Xs.dtype).max
        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        # we are given l <= e <= u, which we translate into e/lf <= e <= e*uf
        # - if the factor is infinite, we limit it to fmax
        # - if the factor is NaN, e.g. from 0/0, we set the factor to 1
        # finally we split the factor geometrically in two using the sqrt
        expr_abs_lower_factor: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.divide(exprv_abs, expr_abs_lower)
        )
        expr_abs_lower_factor[np.isinf(expr_abs_lower_factor)] = fmax
        np.sqrt(expr_abs_lower_factor, out=expr_abs_lower_factor)
        expr_abs_lower_factor[np.isnan(expr_abs_lower_factor)] = 1

        expr_abs_upper_factor: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.divide(
                expr_abs_upper,
                # we avoid division by zero here
                # - if exprv_abs = 0 and expr_abs_upper = 0, factor = 0 works
                # - if exprv_abs = 0 and expr_abs_upper > 0, factor is large
                # - if exprv_abs > 0 and expr_abs_upper > 0, works as normal
                _maximum_zero_sign_sensitive(exprv_abs, smallest_subnormal),
            )
        )
        expr_abs_upper_factor[np.isinf(expr_abs_upper_factor)] = fmax
        np.sqrt(expr_abs_upper_factor, out=expr_abs_upper_factor)
        expr_abs_upper_factor[np.isnan(expr_abs_upper_factor)] = 1

        # TODO: for infinite a*b that includes finite bounds, allow them
        # TODO: we could peek at a and b if they're powers with a constant
        #       exponent to weigh the factor split
        # TODO: handle all terms of a left-associative product in one go

        # compute the bounds for abs(a) and abs(b)
        # - a lower bound of zero is always passed through
        # - we ensure that the upper bound does not overflow into inf
        a_abs_lower = _ensure_array(np.divide(av_abs, expr_abs_lower_factor))
        a_abs_lower[(expr_abs_lower == 0) & ~np.isnan(av_abs)] = 0
        a_abs_upper = _ensure_array(np.multiply(av_abs, expr_abs_upper_factor))
        a_abs_upper[np.isinf(a_abs_upper) & ~np.isinf(expr_abs_upper)] = fmax

        b_abs_lower = _ensure_array(np.divide(bv_abs, expr_abs_lower_factor))
        b_abs_lower[(expr_abs_lower == 0) & ~np.isnan(bv_abs)] = 0
        b_abs_upper = _ensure_array(np.multiply(bv_abs, expr_abs_upper_factor))
        b_abs_upper[np.isinf(b_abs_upper) & ~np.isinf(expr_abs_upper)] = fmax

        any_nan = np.isnan(av_abs)
        any_nan |= np.isnan(bv_abs)
        any_zero = a_abs_lower == 0
        any_zero |= b_abs_lower == 0
        any_inf = np.isinf(a_abs_upper)
        any_inf |= np.isinf(b_abs_upper)
        zero_inf_clash = any_zero & any_inf

        # we cannot allow 0*inf, so truncate the bounds to not include them
        # if any term is NaN, other non-NaN terms can have any value
        a_abs_lower[zero_inf_clash & (a_abs_lower == 0)] = smallest_subnormal
        a_abs_upper[zero_inf_clash & np.isinf(a_abs_upper)] = fmax
        a_abs_lower[any_nan & ~np.isnan(av_abs)] = 0
        a_abs_upper[any_nan & ~np.isnan(av_abs)] = np.inf

        b_abs_lower[zero_inf_clash & (b_abs_lower == 0)] = smallest_subnormal
        b_abs_upper[zero_inf_clash & np.isinf(b_abs_upper)] = fmax
        b_abs_lower[any_nan & ~np.isnan(bv_abs)] = 0
        b_abs_upper[any_nan & ~np.isnan(bv_abs)] = np.inf

        # ensure that the bounds on abs(a) and abs(b) include their values
        a_abs_lower = _ensure_array(_minimum_zero_sign_sensitive(av_abs, a_abs_lower))
        a_abs_upper = _ensure_array(_maximum_zero_sign_sensitive(av_abs, a_abs_upper))

        b_abs_lower = _ensure_array(_minimum_zero_sign_sensitive(bv_abs, b_abs_lower))
        b_abs_upper = _ensure_array(_maximum_zero_sign_sensitive(bv_abs, b_abs_upper))

        # stack the bounds on a and b so that we can nudge their bounds, if
        #  necessary, together
        tl_abs_stack = _stack([a_abs_lower, b_abs_lower])
        tu_abs_stack = _stack([a_abs_upper, b_abs_upper])

        def compute_term_product(
            t_stack: np.ndarray[tuple[int, Ps], np.dtype[F]],
        ) -> np.ndarray[tuple[int, Ps], np.dtype[F]]:
            total_product: np.ndarray[tuple[Ps], np.dtype[F]] = np.multiply(
                t_stack[0], t_stack[1]
            )

            return _broadcast_to(
                _ensure_array(total_product).reshape((1, *exprv_abs.shape)),
                (t_stack.shape[0], *exprv_abs.shape),
            )

        tl_abs_stack = guarantee_stacked_arg_within_expr_bounds(
            compute_term_product,
            _broadcast_to(
                exprv_abs.reshape((1, *exprv_abs.shape)),
                (tl_abs_stack.shape[0], *exprv_abs.shape),
            ),
            _stack([av_abs, bv_abs]),
            tl_abs_stack,
            _broadcast_to(
                expr_abs_lower.reshape((1, *exprv_abs.shape)),
                (tl_abs_stack.shape[0], *exprv_abs.shape),
            ),
            _broadcast_to(
                expr_abs_upper.reshape((1, *exprv_abs.shape)),
                (tl_abs_stack.shape[0], *exprv_abs.shape),
            ),
        )
        tu_abs_stack = guarantee_stacked_arg_within_expr_bounds(
            compute_term_product,
            _broadcast_to(
                exprv_abs.reshape((1, *exprv_abs.shape)),
                (tu_abs_stack.shape[0], *exprv_abs.shape),
            ),
            _stack([av_abs, bv_abs]),
            tu_abs_stack,
            _broadcast_to(
                expr_abs_lower.reshape((1, *exprv_abs.shape)),
                (tu_abs_stack.shape[0], *exprv_abs.shape),
            ),
            _broadcast_to(
                expr_abs_upper.reshape((1, *exprv_abs.shape)),
                (tu_abs_stack.shape[0], *exprv_abs.shape),
            ),
        )

        # derive the bounds on a and b based on the bounds on abs(a) and abs(b)
        # if any term is NaN, other non-NaN terms can have any value of any sign
        a_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _where(_is_sign_negative_number(av), -tu_abs_stack[0], tl_abs_stack[0])
        )
        a_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _where(_is_sign_negative_number(av), -tl_abs_stack[0], tu_abs_stack[0])
        )
        a_lower[any_nan & ~np.isnan(av)] = -np.inf
        a_upper[any_nan & ~np.isnan(av)] = np.inf

        b_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _where(_is_sign_negative_number(bv), -tu_abs_stack[1], tl_abs_stack[1])
        )
        b_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _where(_is_sign_negative_number(bv), -tl_abs_stack[1], tu_abs_stack[1])
        )
        b_lower[any_nan & ~np.isnan(bv)] = -np.inf
        b_upper[any_nan & ~np.isnan(bv)] = np.inf

        # recurse into a and b to propagate their bounds, then combine their
        #  bounds on Xs
        Xs_lower, Xs_upper = a.compute_data_bounds(
            a_lower,
            a_upper,
            Xs,
            late_bound,
        )

        bl, bu = b.compute_data_bounds(
            b_lower,
            b_upper,
            Xs,
            late_bound,
        )
        Xs_lower = _maximum_zero_sign_sensitive(Xs_lower, bl)
        Xs_upper = _minimum_zero_sign_sensitive(Xs_upper, bu)

        # ensure that the bounds on Xs include Xs
        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} * {self._b!r}"


class ScalarDivide(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __new__(cls, a: AnyExpr, b: AnyExpr) -> "ScalarDivide | Number":  # type: ignore[misc]
        if isinstance(a, Number) and isinstance(b, Number):
            # symbolical constant propagation for some cases of int / int
            # division always produces a floating-point number
            ai, bi = a.as_int(), b.as_int()
            if (ai is not None) and (bi is not None):
                d = gcd(ai, bi)
                if ai < 0 and bi < 0 and d > 0:
                    # symbolic reduction of -a / -b to a / b
                    d = -d
                if d != 0:
                    # symbolic reduction of (a*d) / (b*d) to a / b
                    assert (ai % d == 0) and (bi % d == 0)
                    ai //= d
                    bi //= d
                # int / 1 has an exact floating-point result
                if bi == 1:
                    return Number.from_symbolic_int_as_float(ai)
                # int / -1 has an exact floating-point result
                if bi == -1:
                    assert ai >= 0
                    # ensure that 0/1 = 0.0 and 0/-1 = -0.0
                    return Number.from_symbolic_int_as_float(ai, force_negative=True)
                # keep a / b after reduction
                if d != 0:
                    a = Number.from_symbolic_int(ai)
                    b = Number.from_symbolic_int(bi)
        # we have to also assign inside __new__ since the symbolic constant
        #  folding check can change a and b to simplify the expression
        this = super().__new__(cls)
        this._a = a
        this._b = b
        return this

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarDivide | Number":
        return ScalarDivide(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a, self._b, dtype, np.divide, ScalarDivide
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.divide(self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a_const = not self._a.has_data
        b_const = not self._b.has_data
        assert not (a_const and b_const), "constant division has no data bounds"

        # evaluate a and b and a*b
        a, b = self._a, self._b
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)
        exprv = np.divide(av, bv)

        fmax = np.finfo(Xs.dtype).max
        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        term_lower: np.ndarray[tuple[Ps], np.dtype[F]]
        term_upper: np.ndarray[tuple[Ps], np.dtype[F]]

        if a_const:
            term, termv, constv = b, bv, av

            expr_lower, expr_upper = (
                _ensure_array(expr_lower, copy=True),
                _ensure_array(expr_upper, copy=True),
            )

            # ensure that the expression keeps the same sign
            expr_lower[(expr_lower <= 0) & _is_sign_positive_number(exprv)] = +0.0
            expr_upper[(expr_upper >= 0) & _is_sign_negative_number(exprv)] = -0.0

            # compute the divisor bounds
            # for Inf/x, we can allow any finite x with the same sign
            # for 0/x, we can allow any non-zero non-NaN x with the same sign
            # for NaN/x, we can allow any x but only propagate [-inf, inf]
            #  since [-NaN, NaN] would be misunderstood as only NaN
            # otherwise ensure that the divisor keeps the same sign:
            #  - c < 0, t >= +0: el <= e <= eu <= -0 -> tl = el, tu = eu
            #  - c < 0, t <= -0: +0 <= el <= e <= eu -> tl = el, tu = eu
            #  - c > 0, t >= +0: +0 <= el <= e <= eu -> tl = eu, tu = el
            #  - c > 0, t <= -0: el <= e <= eu <= -0 -> tl = eu, tu = el
            # if term_lower == termv and termv == -0.0, we need to guarantee
            #  that term_lower is also -0.0, same for term_upper
            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future
            term_lower = _ensure_array(expr_upper, copy=True)
            np.copyto(
                term_lower,
                expr_lower,
                where=_is_sign_negative_number(constv),
                casting="no",
            )
            np.divide(constv, term_lower, out=term_lower)
            term_lower[np.isnan(constv)] = -np.inf
            term_lower[constv == 0] = smallest_subnormal
            term_lower[(constv == 0) & _is_sign_negative_number(termv)] = -np.inf
            np.copyto(
                term_lower, termv, where=((constv == 0) & (termv == 0)), casting="no"
            )
            term_lower[np.isinf(constv)] = +0.0
            term_lower[np.isinf(constv) & _is_sign_negative_number(termv)] = -fmax
            term_lower = _ensure_array(_minimum_zero_sign_sensitive(termv, term_lower))

            term_upper = _ensure_array(expr_lower, copy=True)
            np.copyto(
                term_upper,
                expr_upper,
                where=_is_sign_negative_number(constv),
                casting="no",
            )
            np.divide(constv, term_upper, out=term_upper)
            term_upper[np.isnan(constv)] = np.inf
            term_upper[constv == 0] = np.inf
            term_upper[
                (constv == 0) & _is_sign_negative_number(termv)
            ] = -smallest_subnormal
            np.copyto(
                term_upper, termv, where=((constv == 0) & (termv == 0)), casting="no"
            )
            term_upper[np.isinf(constv)] = fmax
            term_upper[np.isinf(constv) & _is_sign_negative_number(termv)] = -0.0
            term_upper = _ensure_array(_maximum_zero_sign_sensitive(termv, term_upper))

            # we need to force termv if expr_lower == expr_upper
            np.copyto(term_lower, termv, where=(expr_lower == expr_upper), casting="no")
            np.copyto(term_upper, termv, where=(expr_lower == expr_upper), casting="no")

            # handle rounding errors in divide(divide(...)) early
            term_lower = guarantee_arg_within_expr_bounds(
                lambda term_lower: np.divide(constv, term_lower),
                exprv,
                termv,
                term_lower,
                expr_lower,
                expr_upper,
            )
            term_upper = guarantee_arg_within_expr_bounds(
                lambda term_upper: np.divide(constv, term_upper),
                exprv,
                termv,
                term_upper,
                expr_lower,
                expr_upper,
            )

            return term.compute_data_bounds(
                term_lower,
                term_upper,
                Xs,
                late_bound,
            )

        if b_const:
            term, termv, constv = a, av, bv

            # for x/Inf, we can allow any finite x, unless the output +-0.0 sign
            #  is restricted, then we need to restrict the sign of x:
            #   | constv | expr_lower | expr_upper | term_lower | term_upper |
            #   | +Inf   | >= +0.0    | >= +0.0    | +0.0       | +fmax      |
            #   | +Inf   | <= -0.0    | <= -0.0    | -fmax      | -0.0       |
            #   | +Inf   | <= -0.0    | >= +0.0    | -fmax      | +fmax      |
            #   | -Inf   | >= +0.0    | >= +0.0    | -fmax      | -0.0       |
            #   | -Inf   | <= -0.0    | <= -0.0    | +0.0       | +fmax      |
            #   | -Inf   | <= -0.0    | >= +0.0    | -fmax      | +fmax      |
            #  - term_lower := +0.0 if
            #    signbit(constv) == signbit(expr_lower) == signbit(expr_upper)
            #  - term_upper := -0.0 if
            #    ~signbit(constv) == signbit(expr_lower) == signbit(expr_upper)
            # for x/0, we can allow any non-zero non-NaN x with the same sign
            # for x/NaN, we can allow any x but only propagate [-inf, inf]
            #  since [-NaN, NaN] would be misunderstood as only NaN
            # if term_lower == termv and termv == -0.0, we need to guarantee
            #  that term_lower is also -0.0, same for term_upper
            term_lower = _ensure_array(expr_lower, copy=True)
            np.copyto(
                term_lower,
                expr_upper,
                where=_is_sign_negative_number(constv),
                casting="no",
            )
            np.multiply(term_lower, constv, out=term_lower)
            term_lower[np.isnan(constv)] = -np.inf
            term_lower[constv == 0] = smallest_subnormal
            term_lower[(constv == 0) & _is_sign_negative_number(termv)] = -np.inf
            np.copyto(
                term_lower, termv, where=((constv == 0) & (termv == 0)), casting="no"
            )
            term_lower[np.isinf(constv)] = -fmax
            term_lower[
                np.isinf(constv)
                & (
                    _is_sign_positive_number(constv)
                    == _is_sign_positive_number(expr_lower)
                )
                & (
                    _is_sign_positive_number(constv)
                    == _is_sign_positive_number(expr_upper)
                )
            ] = +0.0
            term_lower = _ensure_array(_minimum_zero_sign_sensitive(termv, term_lower))

            term_upper = _ensure_array(expr_upper, copy=True)
            np.copyto(
                term_upper,
                expr_lower,
                where=_is_sign_negative_number(constv),
                casting="no",
            )
            np.multiply(term_upper, constv, out=term_upper)
            term_upper[np.isnan(constv)] = np.inf
            term_upper[constv == 0] = np.inf
            term_upper[
                (constv == 0) & _is_sign_negative_number(termv)
            ] = -smallest_subnormal
            np.copyto(
                term_upper, termv, where=((constv == 0) & (termv == 0)), casting="no"
            )
            term_upper[np.isinf(constv)] = fmax
            term_upper[
                np.isinf(constv)
                & (
                    _is_sign_negative_number(constv)
                    != _is_sign_negative_number(expr_lower)
                )
                & (
                    _is_sign_negative_number(constv)
                    != _is_sign_negative_number(expr_upper)
                )
            ] = -0.0
            term_upper = _ensure_array(_maximum_zero_sign_sensitive(termv, term_upper))

            # we need to force termv if expr_lower == expr_upper and constv is
            #  finite non-zero (in other cases we explicitly expand ranges)
            np.copyto(
                term_lower,
                termv,
                where=(
                    (expr_lower == expr_upper) & np.isfinite(constv) & (constv != 0)
                ),
                casting="no",
            )
            np.copyto(
                term_upper,
                termv,
                where=(
                    (expr_lower == expr_upper) & np.isfinite(constv) & (constv != 0)
                ),
                casting="no",
            )

            # handle rounding errors in divide(multiply(...)) early
            term_lower = guarantee_arg_within_expr_bounds(
                lambda term_lower: np.divide(term_lower, constv),
                exprv,
                termv,
                term_lower,
                expr_lower,
                expr_upper,
            )
            term_upper = guarantee_arg_within_expr_bounds(
                lambda term_upper: np.divide(term_upper, constv),
                exprv,
                termv,
                term_upper,
                expr_lower,
                expr_upper,
            )

            return term.compute_data_bounds(
                term_lower,
                term_upper,
                Xs,
                late_bound,
            )

        # if neither a not b is const, we simplify by not allowing the sign of
        #  a/b to change
        # we further simplify the code by working with abs(a/b) = abs(a)/abs(b)
        #  and only taking the sign of a and b into account at the end

        # limit expr_lower and expr_upper to keep the same sign
        # then compute the lower and upper bounds for the abs(expr)
        expr_lower = _ensure_array(expr_lower, copy=True)
        expr_lower[_is_sign_positive_number(exprv) & (expr_lower <= 0)] = +0.0
        expr_upper = _ensure_array(expr_upper, copy=True)
        expr_upper[_is_sign_negative_number(exprv) & (expr_upper >= 0)] = -0.0
        expr_abs_lower, expr_abs_upper = (
            _ensure_array(
                _where(_is_sign_negative_number(exprv), -expr_upper, expr_lower)
            ),
            _ensure_array(
                _where(_is_sign_negative_number(exprv), -expr_lower, expr_upper)
            ),
        )

        av_abs = np.abs(av)
        bv_abs = np.abs(bv)
        exprv_abs = _ensure_array(np.abs(exprv))

        fmax = np.finfo(Xs.dtype).max
        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        # we are given l <= e <= u, which we translate into e/lf <= e <= e*uf
        # - if the factor is infinite, we limit it to fmax
        # - if the factor is NaN, e.g. from 0/0, we set the factor to 1
        # finally we split the factor geometrically in two using the sqrt
        expr_abs_lower_factor: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.divide(exprv_abs, expr_abs_lower)
        )
        expr_abs_lower_factor[np.isinf(expr_abs_lower_factor)] = fmax
        np.sqrt(expr_abs_lower_factor, out=expr_abs_lower_factor)
        expr_abs_lower_factor[np.isnan(expr_abs_lower_factor)] = 1

        expr_abs_upper_factor: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.divide(
                expr_abs_upper,
                # we avoid division by zero here
                _maximum_zero_sign_sensitive(exprv_abs, smallest_subnormal),
            )
        )
        expr_abs_upper_factor[np.isinf(expr_abs_upper_factor)] = fmax
        np.sqrt(expr_abs_upper_factor, out=expr_abs_upper_factor)
        expr_abs_upper_factor[np.isnan(expr_abs_upper_factor)] = 1

        # TODO: for infinite a/b that includes finite bounds, allow them
        # TODO: we could peek at a and b if they're powers with a constant
        #       exponent to weigh the factor split
        # TODO: handle all terms of a left-associative product in one go

        # compute the bounds for abs(a) and abs(b)
        # - a lower bound of zero is always passed through
        # - we ensure that the upper bound does not overflow into inf
        # - the bounds for the divisor b are flipped
        a_abs_lower = _ensure_array(np.divide(av_abs, expr_abs_lower_factor))
        a_abs_lower[(expr_abs_lower == 0) & ~np.isnan(av_abs)] = 0
        a_abs_upper = _ensure_array(np.multiply(av_abs, expr_abs_upper_factor))
        a_abs_upper[np.isinf(a_abs_upper) & ~np.isinf(expr_abs_upper)] = fmax

        b_abs_lower = _ensure_array(np.multiply(bv_abs, expr_abs_lower_factor))
        b_abs_lower[(expr_abs_lower == 0) & ~np.isnan(av_abs)] = np.inf
        b_abs_upper = _ensure_array(np.divide(bv_abs, expr_abs_upper_factor))
        b_abs_upper[(b_abs_upper == 0) & ~np.isinf(expr_abs_upper)] = smallest_subnormal

        any_nan = np.isnan(av_abs)
        any_nan |= np.isnan(bv_abs)
        both_zero = a_abs_lower == 0
        both_zero &= b_abs_lower == 0
        both_inf = np.isinf(a_abs_upper)
        both_inf &= np.isinf(b_abs_upper)
        zero_inf_clash = both_zero | both_inf

        # we cannot allow 0/0 or inf/inf, so truncate the bounds to not include
        #  them
        # if any term is NaN, other non-NaN terms can have any value
        a_abs_lower[zero_inf_clash & (a_abs_lower == 0)] = smallest_subnormal
        a_abs_upper[zero_inf_clash & np.isinf(a_abs_upper)] = fmax
        a_abs_lower[any_nan & ~np.isnan(av_abs)] = 0
        a_abs_upper[any_nan & ~np.isnan(av_abs)] = np.inf

        b_abs_lower[zero_inf_clash & (b_abs_lower == 0)] = smallest_subnormal
        b_abs_upper[zero_inf_clash & np.isinf(b_abs_upper)] = fmax
        b_abs_lower[any_nan & ~np.isnan(bv_abs)] = 0
        b_abs_upper[any_nan & ~np.isnan(bv_abs)] = np.inf

        # ensure that the bounds on abs(a) and abs(b) include their values
        a_abs_lower = _ensure_array(_minimum_zero_sign_sensitive(av_abs, a_abs_lower))
        a_abs_upper = _ensure_array(_maximum_zero_sign_sensitive(av_abs, a_abs_upper))

        b_abs_lower = _ensure_array(_minimum_zero_sign_sensitive(bv_abs, b_abs_lower))
        b_abs_upper = _ensure_array(_maximum_zero_sign_sensitive(bv_abs, b_abs_upper))

        # stack the bounds on a and b so that we can nudge their bounds, if
        #  necessary, together
        tl_abs_stack = _stack([a_abs_lower, b_abs_lower])
        tu_abs_stack = _stack([a_abs_upper, b_abs_upper])

        def compute_term_product(
            t_stack: np.ndarray[tuple[int, Ps], np.dtype[F]],
        ) -> np.ndarray[tuple[int, Ps], np.dtype[F]]:
            total_product: np.ndarray[tuple[Ps], np.dtype[F]] = np.divide(
                t_stack[0], t_stack[1]
            )

            return _broadcast_to(
                _ensure_array(total_product).reshape((1, *exprv_abs.shape)),
                (t_stack.shape[0], *exprv_abs.shape),
            )

        tl_abs_stack = guarantee_stacked_arg_within_expr_bounds(
            compute_term_product,
            _broadcast_to(
                exprv_abs.reshape((1, *exprv_abs.shape)),
                (tl_abs_stack.shape[0], *exprv_abs.shape),
            ),
            _stack([av_abs, bv_abs]),
            tl_abs_stack,
            _broadcast_to(
                expr_abs_lower.reshape((1, *exprv_abs.shape)),
                (tl_abs_stack.shape[0], *exprv_abs.shape),
            ),
            _broadcast_to(
                expr_abs_upper.reshape((1, *exprv_abs.shape)),
                (tl_abs_stack.shape[0], *exprv_abs.shape),
            ),
        )
        tu_abs_stack = guarantee_stacked_arg_within_expr_bounds(
            compute_term_product,
            _broadcast_to(
                exprv_abs.reshape((1, *exprv_abs.shape)),
                (tu_abs_stack.shape[0], *exprv_abs.shape),
            ),
            _stack([av_abs, bv_abs]),
            tu_abs_stack,
            _broadcast_to(
                expr_abs_lower.reshape((1, *exprv_abs.shape)),
                (tu_abs_stack.shape[0], *exprv_abs.shape),
            ),
            _broadcast_to(
                expr_abs_upper.reshape((1, *exprv_abs.shape)),
                (tu_abs_stack.shape[0], *exprv_abs.shape),
            ),
        )

        # derive the bounds on a and b based on the bounds on abs(a) and abs(b)
        # if any term is NaN, other non-NaN terms can have any value of any sign
        # the bounds for the divisor b are flipped back here
        a_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _where(_is_sign_negative_number(av), -tu_abs_stack[0], tl_abs_stack[0])
        )
        a_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _where(_is_sign_negative_number(av), -tl_abs_stack[0], tu_abs_stack[0])
        )
        a_lower[any_nan & ~np.isnan(av)] = -np.inf
        a_upper[any_nan & ~np.isnan(av)] = np.inf

        b_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _minimum_zero_sign_sensitive(
            tl_abs_stack[1], tu_abs_stack[1]
        )
        b_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _maximum_zero_sign_sensitive(
            tl_abs_stack[1], tu_abs_stack[1]
        )
        b_lower, b_upper = (
            _ensure_array(_where(_is_sign_negative_number(bv), -b_upper, b_lower)),
            _ensure_array(_where(_is_sign_negative_number(bv), -b_lower, b_upper)),
        )
        b_lower[any_nan & ~np.isnan(bv)] = -np.inf
        b_upper[any_nan & ~np.isnan(bv)] = np.inf

        # recurse into a and b to propagate their bounds, then combine their
        #  bounds on Xs
        Xs_lower, Xs_upper = a.compute_data_bounds(
            a_lower,
            a_upper,
            Xs,
            late_bound,
        )

        bl, bu = b.compute_data_bounds(
            b_lower,
            b_upper,
            Xs,
            late_bound,
        )
        Xs_lower = _maximum_zero_sign_sensitive(Xs_lower, bl)
        Xs_upper = _minimum_zero_sign_sensitive(Xs_upper, bu)

        # ensure that the bounds on Xs include Xs
        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} / {self._b!r}"



"""
Commonly used type variables.
"""

__all__ = ["C", "F", "Fi", "S", "Si", "T", "Ti", "TB", "U", "JSON"]

from typing import TypeAlias, TypeVar

import numpy as np

C = TypeVar("C", bound=np.unsignedinteger, covariant=True)
""" The numpy data type for safeguard corrections (covariant). """

F = TypeVar("F", bound=np.floating, covariant=True)
""" Any numpy [`floating`][numpy.floating]-point data type (covariant). """

Fi = TypeVar("Fi", bound=np.floating)
""" Any numpy [`floating`][numpy.floating]-point data type (invariant). """

S = TypeVar("S", bound=tuple[int, ...], covariant=True)
""" Any array shape (covariant). """

Si = TypeVar("Si", bound=tuple[int, ...])
""" Any array shape (invariant). """

T = TypeVar("T", bound=np.number, covariant=True)
""" Any numpy [`number`][numpy.number] data type (covariant). """

Ti = TypeVar("Ti", bound=np.number)
""" Any numpy [`number`][numpy.number] data type (invariant). """

TB = TypeVar("TB", bound=np.number | np.bool, covariant=True)
""" Any numpy [`number`][numpy.number] or [`boolean`][numpy.bool] data type (covariant). """

U = TypeVar("U", bound=np.unsignedinteger, covariant=True)
""" Any numpy [`unsignedinteger`][numpy.unsignedinteger] data type (covariant). """

JSON: TypeAlias = None | int | float | str | bool | list["JSON"] | dict[str, "JSON"]
""" Types that are valid JSON and can be encoded with [`json.dumps`][json.dumps]. """



from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _is_sign_negative_number,
    _is_sign_positive_number,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds, guarantee_arg_within_expr_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarReciprocal(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarReciprocal":
        return ScalarReciprocal(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.reciprocal,
            ScalarReciprocal,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.reciprocal(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and reciprocal(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.reciprocal(argv)

        # compute the argument bounds
        # ensure that reciprocal(...) keeps the same sign as arg
        # TODO: an interval union could represent that the two disjoint
        #       intervals in the future
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _minimum_zero_sign_sensitive(expr_upper, Xs.dtype.type(-0.0))
        )
        np.copyto(
            arg_lower, expr_upper, where=_is_sign_positive_number(exprv), casting="no"
        )
        np.reciprocal(arg_lower, out=arg_lower)
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _maximum_zero_sign_sensitive(Xs.dtype.type(+0.0), expr_lower)
        )
        np.copyto(
            arg_upper, expr_lower, where=_is_sign_negative_number(exprv), casting="no"
        )
        np.reciprocal(arg_upper, out=arg_upper)
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in reciprocal(reciprocal(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.reciprocal(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.reciprocal(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"reciprocal({self._a!r})"



"""
Implementations for the provided [`StencilSafeguard`][compression_safeguards.safeguards.stencil.abc.StencilSafeguard]s.
"""

__all__ = ["BoundaryCondition", "NeighbourhoodAxis", "NeighbourhoodBoundaryAxis"]

from enum import Enum, auto
from functools import reduce
from typing import Literal, Self, assert_never

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils._compat import _sliding_window_view
from ...utils.bindings import Parameter
from ...utils.error import TypeCheckError, ctx, lookup_enum_or_raise
from ...utils.typing import JSON, TB, S


class BoundaryCondition(Enum):
    """
    Different types of boundary conditions that can be applied to the data
    array domain boundaries for [`StencilSafeguard`][..abc.StencilSafeguard]s.

    Since stencil safeguards operate over small neighbourhoods of data points,
    points at the boundary, where part of the neighbourhood may not exist, need
    to be treated specially.
    """

    valid = auto()
    """
    The boundary is not extended, instead the safeguard is only applied to
    and checked for points where the entire neighbourhood is valid.
    """

    constant = auto()
    """
    The boundary is extended by a constant value.
    """

    edge = auto()
    """
    The boundary is extended by the edge value.
    """

    reflect = auto()
    """
    The boundary is extended by reflecting along the edge value. The edge value
    itself is not repeated.
    """

    symmetric = auto()
    """
    The boundary is extended by reflecting after the edge value. The edge value
    itself is repeated as well.
    """

    wrap = auto()
    """
    The boundary is extended by wrapping the domain around, as if the domain was
    on a torus (Pac-Man style).
    """


class NeighbourhoodAxis:
    """
    Specification of the shape of the data neighbourhood along a single axis.

    Parameters
    ----------
    before : int
        The non-negative number of values to include before the centre of a
        data neighbourhood.

        e.g. setting `before=1` means that the neighbourhood contains the
        previous value.
    after : int
        The non-negative number of values to include after the centre of a
        data neighbourhood.

        e.g. setting `after=2` means that the neighbourhood contains the
        two next values.

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    ValueError
        if `before` or `after` is negative.
    """

    __slots__: tuple[str, ...] = ("_before", "_after")
    _before: int
    _after: int

    def __init__(
        self,
        before: int,
        after: int,
    ) -> None:
        with ctx.parameter("before"):
            TypeCheckError.check_instance_or_raise(before, int)
            if before < 0:
                raise ValueError("must be non-negative") | ctx
            self._before = before

        with ctx.parameter("after"):
            TypeCheckError.check_instance_or_raise(after, int)
            if after < 0:
                raise ValueError("must be non-negative") | ctx
            self._after = after

    @property
    def before(self) -> int:
        """
        The non-negative number of values to include before the centre of a
        data neighbourhood.
        """
        return self._before

    @property
    def after(self) -> int:
        """
        The non-negative number of values to include after the centre of a
        data neighbourhood.
        """
        return self._after

    @override
    def __repr__(self) -> str:
        return f"{type(self).__name__}(before={self.before}, after={self.after})"


class NeighbourhoodBoundaryAxis:
    """
    Specification of the shape of the data neighbourhood and its boundary
    condition along a single axis.

    Parameters
    ----------
    axis : int
        The axis along which the boundary condition is applied.
    before : int
        The non-negative number of values to include before the centre of a
        data neighbourhood.

        e.g. setting `before=1` means that the neighbourhood contains the
        previous value.
    after : int
        The non-negative number of values to include after the centre of a
        data neighbourhood.

        e.g. setting `after=2` means that the neighbourhood contains the
        two next values.
    boundary : str | BoundaryCondition
        The boundary condition that is applied to this axis near the data
        array domain boundary to fill the data neighbourhood, e.g. by extending
        values.
    constant_boundary : None | int | float | str | Parameter
        The optional value of or the late-bound parameter name for the constant
        value with which the data array domain is extended for a constant
        boundary. The value must be losslessly convertible to the data dtype.

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    ValueError
        if `before` or `after` is negative.
    ValueError
        if `boundary` does not name a valid boundary condition variant.
    ValueError
        if `constant_boundary` is, not, provided if and only if the `boundary`
        is constant.
    ValueError
        if `constant_boundary` uses the non-scalar `$x` or `$X` late-bound
        parameters.
    """

    __slots__: tuple[str, ...] = ("_axis", "_shape", "_boundary", "_constant_boundary")
    _axis: int
    _shape: NeighbourhoodAxis
    _boundary: BoundaryCondition
    _constant_boundary: None | int | float | Parameter

    def __init__(
        self,
        axis: int,
        before: int,
        after: int,
        boundary: str | BoundaryCondition,
        constant_boundary: None | int | float | str | Parameter = None,
    ) -> None:
        with ctx.parameter("axis"):
            TypeCheckError.check_instance_or_raise(axis, int)
            self._axis = axis

        self._shape = NeighbourhoodAxis(before, after)

        with ctx.parameter("boundary"):
            TypeCheckError.check_instance_or_raise(boundary, str | BoundaryCondition)
            self._boundary = (
                boundary
                if isinstance(boundary, BoundaryCondition)
                else lookup_enum_or_raise(BoundaryCondition, boundary)
            )

        with ctx.parameter("constant_boundary"):
            TypeCheckError.check_instance_or_raise(
                constant_boundary, None | int | float | str | Parameter
            )

            if (self._boundary != BoundaryCondition.constant) != (
                constant_boundary is None
            ):
                raise (
                    ValueError(
                        "must be provided if and only if the constant "
                        + "boundary condition is used"
                    )
                    | ctx
                )

            if isinstance(constant_boundary, Parameter):
                self._constant_boundary = constant_boundary
            elif isinstance(constant_boundary, str):
                self._constant_boundary = Parameter(constant_boundary)
            else:
                self._constant_boundary = constant_boundary

            if isinstance(self._constant_boundary, Parameter):
                if self._constant_boundary in ["$x", "$X"]:
                    raise (
                        ValueError(
                            "must be a scalar but late-bound constant data "
                            + f"{self._constant_boundary} may not be"
                        )
                        | ctx
                    )

    @property
    def axis(self) -> int:
        """
        The axis along which the boundary condition is applied.
        """
        return self._axis

    @property
    def before(self) -> int:
        """
        The non-negative number of values to include before the centre of a
        data neighbourhood.
        """
        return self._shape.before

    @property
    def after(self) -> int:
        """
        The non-negative number of values to include after the centre of a
        data neighbourhood.
        """
        return self._shape.after

    @property
    def shape(self) -> NeighbourhoodAxis:
        """
        The shape of the data neighbourhood.
        """
        return self._shape

    @property
    def boundary(self) -> BoundaryCondition:
        """
        The boundary condition that is applied to this axis near the data
        array domain boundary to fill the data neighbourhood, e.g. by extending
        values.
        """
        return self._boundary

    @property
    def constant_boundary(self) -> None | int | float | Parameter:
        """
        The optional value of or the late-bound parameter name for the constant
        value with which the data array domain is extended for a constant
        boundary.
        """
        return self._constant_boundary

    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the data neighbourhood.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the data neighbourhood.
        """

        config: dict[str, JSON] = dict(
            axis=self.axis,
            before=self.before,
            after=self.after,
            boundary=self.boundary.name,
            constant_boundary=str(self.constant_boundary)
            if isinstance(self.constant_boundary, Parameter)
            else self.constant_boundary,
        )

        if self.constant_boundary is None:
            del config["constant_boundary"]

        return config

    @classmethod
    def from_config(cls, config: dict[str, JSON]) -> Self:
        """
        Instantiate the data neighbourhood from a configuration [`dict`][dict].

        Parameters
        ----------
        config : dict[str, JSON]
            Configuration of the data neighbourhood.

        Returns
        -------
        neighbourhood : Self
            Instantiated data neighbourhood.
        """

        return cls(**config)  # type: ignore

    @override
    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(f'{k}={v!r}' for k, v in self.get_config().items())})"


def _pad_with_boundary(
    a: np.ndarray[S, np.dtype[TB]],
    boundary: BoundaryCondition,
    pad_before: int,
    pad_after: int,
    constant: None | np.ndarray[tuple[()], np.dtype[TB]],
    axis: int,
) -> np.ndarray[tuple[int, ...], np.dtype[TB]]:
    if (axis >= a.ndim) or (axis < -a.ndim):
        return a

    pad_width = [(0, 0)] * a.ndim
    pad_width[axis] = (pad_before, pad_after)

    kwargs: dict[str, None | str | np.ndarray[tuple[()], np.dtype[TB]]] = dict()
    match boundary:
        case BoundaryCondition.valid:
            return a
        case BoundaryCondition.constant:
            mode = "constant"
            kwargs["constant_values"] = constant
        case BoundaryCondition.edge:
            mode = "edge"
        case BoundaryCondition.reflect:
            mode = "reflect"
            kwargs["reflect_type"] = "even"
        case BoundaryCondition.symmetric:
            mode = "symmetric"
            kwargs["reflect_type"] = "even"
        case BoundaryCondition.wrap:
            mode = "wrap"
        case _:
            assert_never(boundary)

    return np.pad(a, pad_width, mode, **kwargs)  # type: ignore


def _reverse_neighbourhood_indices(
    data_shape: tuple[int, ...],
    neighbourhood: tuple[NeighbourhoodBoundaryAxis, ...],
    window_used: np.ndarray[tuple[int, ...], np.dtype[np.bool]],
    where_flat: Literal[True] | np.ndarray[tuple[int], np.dtype[np.bool]],
) -> np.ndarray[tuple[int, int], np.dtype[np.intp]]:
    data_size = reduce(lambda x, y: x * y, data_shape, 1)

    window = tuple(axis.before + 1 + axis.after for axis in neighbourhood)
    window_size = reduce(lambda x, y: x * y, window, 1)

    # compute how the data indices are distributed into windows
    # i.e. for each derived element, which data does it depend on
    indices_boundary = np.arange(data_size).reshape(data_shape)
    for axis in neighbourhood:
        indices_boundary = _pad_with_boundary(
            indices_boundary,
            axis.boundary,
            axis.before,
            axis.after,
            None if axis.constant_boundary is None else np.full((), data_size),
            axis.axis,
        )
    indices_windows = _sliding_window_view(
        indices_boundary,
        window,
        axis=tuple(axis.axis for axis in neighbourhood),
        writeable=False,
    ).reshape((-1, window_size))

    # compute the reverse: for each data element, which windows is it in
    # i.e. for each data element, which derived elements does it contribute to
    #      and thus which data bounds affect it
    reverse_indices_windows = np.full(
        (data_size, np.sum(window_used.astype(int))), indices_windows.size
    )
    reverse_indices_counter = np.zeros(data_size, dtype=np.intp)
    for i, u in enumerate(window_used.flat):
        # skip window indices that are not used
        if not u:
            continue
        # manual loop to account for potential aliasing:
        # with a wrapping boundary, more than one j for the same window
        #  position j could refer back to the same data element
        for j in range(indices_windows.shape[0]):
            # skip back-contributions from data elements where the safety
            #  requirements are disabled
            if (where_flat is not True) and (not where_flat[j]):
                continue
            idx = indices_windows[j, i]
            if idx != data_size:
                # lazily allocate more to account for all possible edge cases
                if reverse_indices_counter[idx] >= reverse_indices_windows.shape[1]:
                    new_reverse_indices_windows = np.full(
                        (data_size, reverse_indices_windows.shape[1] * 2),
                        indices_windows.size,
                    )
                    new_reverse_indices_windows[
                        :, : reverse_indices_windows.shape[1]
                    ] = reverse_indices_windows
                    reverse_indices_windows = new_reverse_indices_windows
                # update the reverse mapping
                reverse_indices_windows[idx][reverse_indices_counter[idx]] = (
                    j * window_used.size
                ) + i
                reverse_indices_counter[idx] += 1

    return reverse_indices_windows



import numpy as np

from ...utils._compat import (
    _is_sign_negative_number,
    _is_sign_positive_number,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
)
from ...utils.cast import from_float, from_total_order, to_float, to_total_order
from ...utils.intervals import Interval, IntervalUnion, Lower, Upper
from ...utils.typing import F, S, T


def compute_safe_data_lower_upper_interval_union(
    data: np.ndarray[S, np.dtype[T]],
    data_float_lower: np.ndarray[S, np.dtype[F]],
    data_float_upper: np.ndarray[S, np.dtype[F]],
) -> IntervalUnion[T, int, int]:
    """
    Compute the safe interval for the `data` that upholds the provided
    `data_upper` and `data_lower` bounds and preserves finite values and
    whether a value is NaN.

    Parameters
    ----------
    data : np.ndarray[S, np.dtype[T]]
        Data for which to compute the safe interval.
    data_float_lower : np.ndarray[S, np.dtype[F]]
        Pointwise lower bounds for the floating-point version of `data`.
    data_float_upper : np.ndarray[S, np.dtype[F]]
        Pointwise upper bounds for the floating-point version of `data`.

    Returns
    -------
    valid : Interval[T, int]
        Safe interval for the `data` to be within the bounds.
    """

    dataf: np.ndarray[tuple[int], np.dtype[T]] = data.flatten()
    dataf_float_lower: np.ndarray[tuple[int], np.dtype[F]] = data_float_lower.flatten()
    dataf_float_upper: np.ndarray[tuple[int], np.dtype[F]] = data_float_upper.flatten()

    valid = Interval.empty_like(dataf).preserve_inf(dataf).preserve_finite(dataf)

    Lower(
        _maximum_zero_sign_sensitive(
            valid._lower, from_float(dataf_float_lower, dataf.dtype)
        )
    ) <= valid[np.isfinite(dataf)] <= Upper(
        _minimum_zero_sign_sensitive(
            from_float(dataf_float_upper, dataf.dtype), valid._upper
        )
    )

    # correct rounding errors in the lower and upper bound
    # the correction needs to be zero-sign sensitive, e.g. if the upper bound
    #  is -0.0 then the integer valid interval upper bound must -1, not (+)0
    with np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore"):
        valid_float_lower = to_float(valid._lower, ftype=dataf_float_lower.dtype)
        valid_float_upper = to_float(valid._upper, ftype=dataf_float_lower.dtype)
        lower_outside_bound = (valid_float_lower < dataf_float_lower) | (
            _is_sign_negative_number(valid_float_lower)
            & _is_sign_positive_number(dataf_float_lower)
        )
        upper_outside_bound = (valid_float_upper > dataf_float_upper) | (
            _is_sign_positive_number(valid_float_upper)
            & _is_sign_negative_number(dataf_float_upper)
        )

    Lower(
        from_total_order(
            to_total_order(valid._lower) + lower_outside_bound,
            data.dtype,
        )
    ) <= valid[np.isfinite(dataf)] <= Upper(
        from_total_order(
            to_total_order(valid._upper) - upper_outside_bound,
            data.dtype,
        )
    )

    return valid.preserve_any_nan(dataf, equal_nan=True)



"""
## Safeguarding Quantities of Interest

We are often not just interested in data itself, but also in quantities derived
from it. For instance, we might later plot the data logarithm, compute a
derivative, or apply a smoothing kernel. In these cases, we often want to
safeguard not just properties on the data but also on these derived quantities
of interest (QoIs).

The `compression-safeguards` package provides the
[`PointwiseQuantityOfInterestErrorBoundSafeguard`][..pointwise.qoi.eb.PointwiseQuantityOfInterestErrorBoundSafeguard]
and
[`StencilQuantityOfInterestErrorBoundSafeguard`][..stencil.qoi.eb.StencilQuantityOfInterestErrorBoundSafeguard]
safeguards to preserve various [`ErrorBound`][..eb.ErrorBound]s on pointwise[^1]
and stencil[^2] quantities of interest, respectively.

[^1]: A pointwise QoI is computed independently for each data point, taking
    only the value of this data point as input.
[^2]: A stencil QoI is computed for a local neighbourhood or stencil around
    each data point, e.g. to compute a 5x5 2D block mean over the data.

## Grammar

These QoI safeguards are configured with the quantity of interest expression,
which must be given in [`str`][str]ing form using the following EBNF
grammar[^3], where some rules are only available in pointwise or in stencil
QoIs:

[^3]: You can visualise the EBNF grammar at <https://matthijsgroen.github.io/ebnf2railroad/try-yourself.html>.

```ebnf
qoi =
    expr
  | { assignment }, "return", expr, ";"
;

assignment =
    variable, "=", expr, ";"
;

expr =
    number
  | array
  | unary_operator
  | binary_operator
  | binary_comparison
  | array_transpose
  | subexpression
  | constant
  | data
  | late_bound_constant
  | variable
  | array_indexing
  | arithmetic_functions
  | classification_functions
  | logical_combinators
  | array_functions
  | finite_difference
;

number =
    integer
  | float
  | "Inf"  (* infinity *)
  | "NaN"  (* not a number *)
;

integer =
    [ sign ], digit, { digit }
;
float =
    [ sign ], digit, { digit }, [
        ".", digit, { digit }
    ], [
        "e", [ sign ], digit, { digit }
    ]
;

sign =
    "+" | "-"
;
digit =
    "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"
;

array =
    "[", expr, { ",", expr }, [","], "]"
;

unary_operator =
    "+", expr  (* positive / no-op *)
  | "-", expr  (* negation *)
;

binary_operator =
    expr, "+", expr  (* addition *)
  | expr, "-", expr  (* subtraction *)
  | expr, "*", expr  (* multiplication *)
  | expr, "/", expr  (* division *)
  | expr, "**", expr  (* exponentiation / power *)
;

binary_comparison =
    expr, "<", expr  (* 1 if a less than b, 0 otherwise (e.g. with NaN) *)
  | expr, "<=", expr  (* 1 if a less than or equal to b, 0 otherwise (e.g. with NaN) *)
  | expr, "==", expr  (* 1 if a equal to b, 0 otherwise (e.g. with NaN) *)
  | expr, "!=", expr  (* 1 if a not equal to b, 0 otherwise; 1 for NaN != NaN *)
  | expr, ">=", expr  (* 1 if a greater than or equal to b, 0 otherwise (e.g. with NaN) *)
  | expr, ">", expr  (* 1 if a greater than b, 0 otherwise (e.g. with NaN) *)
;

array_transpose =
    expr, ".T"
;

subexpression =
    "(", expr, ")"
;

constant =
    "e"  (* Euler's number *)
  | "pi"  (* pi *)
;

data =
    "x"  (* pointwise data value, x = X[I] *)
  | "X"  (* stencil data neighbourhood, only available in stencil QoIs *)
;

late_bound_constant =
    "c", "[", '"', identifier, '"', "]"  (* late-bound constant pointwise value *)
  | "C", "[", '"', identifier, '"', "]"  (* late-bound constant stencil neighbourhood, only available in stencil QoIs *)
  | "c", "[", '"', "$", identifier, '"', "]"  (* built-in late-bound constant pointwise value *)
  | "C", "[", '"', "$", identifier, '"', "]"  (* built-in late-bound constant stencil neighbourhood, only available in stencil QoIs *)
;

variable =
    "v", "[", '"', identifier, '"', "]"    (* user-defined variable, only available in pointwise QoIs *)
  | "V", "[", '"', identifier, '"', "]"    (* user-defined variable, only available in stencil QoIs *)
;

identifier =
    ( letter | "_" ), { letter | digit | "_" }
;
letter =
    "a" | "b" | "c" | "d" | "e" | "f" | "g" | "h" | "i" | "j" | "k"
  | "l" | "m" | "n" | "o" | "p" | "q" | "r" | "s" | "t" | "u" | "v"
  | "w" | "x" | "y" | "z" | "A" | "B" | "C" | "D" | "E" | "F" | "G"
  | "H" | "I" | "J" | "K" | "L" | "M" | "N" | "O" | "P" | "Q" | "R"
  | "S" | "T" | "U" | "V" | "W" | "X" | "Y" | "Z"
;

array_indexing =
    expr, "[", "I", "]"  (* stencil neighbourhood centre, only available in stencil QoIs *)
  | expr, "[", index, { ",", index }, [","], "]"  (* array indexing *)
  | "I", "[", index, { ",", index }, [","], "]"  (* indexed 1D array over the per-axis stencil neighbourhood centre indices, only available in stencil QoIs *)
;

index =
    integer_expr  (* single index *)
  | [ integer_expr ], ":", [ integer_expr ]  (* slice index from (inclusive) to (exclusive) *)
  | [ integer_expr ], ":", [ integer_expr ], ":", [ integer_expr ]  (* slice index with step *)
;

integer_expr =
    ? symbolic integer-only expression ?
;

arithmetic_functions =
    "ln", "(", expr, [","], ")"  (* natural logarithm *)
  | "log2", "(", expr, [","], ")"  (* binary logarithm *)
  | "log10", "(", expr, [","], ")"  (* decimal logarithm *)
  | "log", "(", expr, ",", "base", "=", expr, [","], ")"  (* logarithm with arbitrary base *)
  | "exp", "(", expr, [","], ")"  (* exponential e^x *)
  | "exp2", "(", expr, [","], ")"  (* binary exponentiation 2^x *)
  | "exp10", "(", expr, [","], ")"  (* decimal exponentiation 10^x *)
  | "sqrt", "(", expr, [","], ")"  (* square root *)
  | "square", "(", expr, [","], ")"  (* square x^2 *)
  | "reciprocal", "(", expr, [","], ")"  (* reciprocal 1/x *)
  | "abs", "(", expr, [","], ")"   (* absolute value *)
  | "sign", "(", expr, [","], ")"  (* sign function, signed NaN for NaNs *)
  | "floor", "(", expr, [","], ")"  (* round down, towards negative infinity *)
  | "ceil", "(", expr, [","], ")"  (* round up, towards positive infinity *)
  | "trunc", "(", expr, [","], ")"  (* round towards zero *)
  | "round_ties_even", "(", expr, [","], ")"  (* round to nearest integer, ties to even *)
  | "sin", "(", expr, [","], ")"  (* sine sin(x) *)
  | "cos", "(", expr, [","], ")"  (* cosine cos(x) *)
  | "tan", "(", expr, [","], ")"  (* tangent tan(x) *)
  | "asin", "(", expr, [","], ")"  (* inverse sine asin(x) *)
  | "acos", "(", expr, [","], ")"  (* inverse cosine acos(x) *)
  | "atan", "(", expr, [","], ")"  (* inverse tangent atan(x) *)
  | "sinh", "(", expr, [","], ")"  (* hyperbolic sine sinh(x) *)
  | "cosh", "(", expr, [","], ")"  (* hyperbolic cosine cosh(x) *)
  | "tanh", "(", expr, [","], ")"  (* hyperbolic tangent tanh(x) *)
  | "asinh", "(", expr, [","], ")"  (* inverse hyperbolic sine asinh(x) *)
  | "acosh", "(", expr, [","], ")"  (* inverse hyperbolic cosine acosh(x) *)
  | "atanh", "(", expr, [","], ")"  (* inverse hyperbolic tangent atanh(x) *)
;

classification_functions =
    "isfinite", "(", expr, [","], ")"  (* 1 if finite, 0 if inf or NaN *)
  | "isinf", "(", expr, [","], ")"  (* 1 if inf, 0 if finite or NaN *)
  | "isnan", "(", expr, [","], ")"  (* 1 if NaN, 0 if finite or inf *)
;

logical_combinators =
    "not", "(", expr, [","], ")"  (* 1 if == 0, 0 if != 0 *)
  | "all", "(", expr, [","], ")"  (* 1 if all array elements != 0, 0 if any array element == 0 *)
  | "any", "(", expr, [","], ")"  (* 1 if any array element != 0, 0 if all array elements == 0 *)
  | "where", "(", expr, ",", expr, ",", expr, [","], ")"  (* where(c, x, y) = x if (c != 0) else y *)
;

array_functions =
    "size", "(", expr, [","], ")"  (* array size *)
  | "shape", "(", expr, [","], ")"  (* array shape as a 1D array *)
  | "sum", "(", expr, [","], ")"  (* array sum *)
  | "matmul", "(", expr, ",", expr, [","], ")"  (* matrix (2d array) multiplication *)
;

finite_difference =
    "finite_difference", "("  (* finite difference over an expression, only available in stencil QoIs *)
      , expr, ","
      , "order", "=", integer, ","  (* order of the derivative *)
      , "accuracy", "=", integer, ","  (* order of accuracy of the approximation *)
      , "type", "=", ("-1" | "0" | "1"), ","  (* backwards | central | forward difference *)
      , "axis", "=", integer, ","  (* axis, relative to the neighbourhood *)
      , (
            "grid_spacing", "=", expr  (* scalar uniform grid spacing along the axis *)
          | "grid_centre", "=", expr  (* centre of an arbitrary grid along the axis *)
        )
      , (
            [","]
          | ",", "grid_period", "=", expr, [","]  (* optional grid period, e.g. 2*pi or 360 *)
        )
  , ")"
;
```

The QoI expression can also contain whitespaces (space ` `, tab `\\t`,
newline `\\n`) and single-line inline comments starting with a hash `#`.

## Numerical Evaluation

### Floating-point data type

QoIs can be evaluated on any data type supported by the safeguards (see
[`Safeguards.supported_dtypes`][...api.Safeguards.supported_dtypes]).
Since the QoIs support many functions with floating-point outputs, they
are evaluated using floating-point arithmetic.

Importantly, the floating-point evaluation data type *must* be able to
represent all values of the input data type losslessly.

* For floating-point data, this is at least the input data type.

* For integer data, the data is first losslessly upcast to a floating-point
type with sufficient precision to represent all integer values, i.e. a type
whose mantissa has more bits than the integer type. For the below floating-
point types, this corresponds to choosing a floating-point data type with a
larger bit width (e.g. at least [`np.float64`][numpy.float64] for
[`np.int32`][numpy.int32] or [`np.uint32`][numpy.uint32] data).

The specific floating-point data type in which the quantities of interest are
evaluated is configured using the [`ToFloatMode`][...utils.cast.ToFloatMode]
enum, please refer to its documentation for further information.

### Literals

The quantities of interest can contain both integer and floating-point
literals. During numerical evaluation, both are evaluated for the chosen
floating-point type. For instance, if evaluation occurs in
[`np.float64`][numpy.float64] format, a literal `0.33` is evaluated using
[`np.float64("0.33")`], which returns the closest representable
[`np.float64`][numpy.float64] value for the symbolic value `0.33`. The
constants `e` and `pi` are provided as the closest representable value in the
chosen floating-point type.

### Symbolic integer constant folding

For integer-only subexpressions, the quantities of interest perform symbolic
constant folding based on operator associativity and explicit grouping with
parentheses. Like in Python, `1 + 2 + 3.5` (addition is left-associative) is
first evaluated to `3 + 3.5` and `3.5 + 2 * 3` to `3.5 + 6`. However, `1.0 + 1`
is left as-is. The following symbolic integer constant-folding operations are
provided:

- group `(a)`: iff `a` is an integer, `(a)` is folded
- negation `-a` (right associative): iff `a` is an integer, `-a` is folded
- addition `a + b` (left associative): iff both `a` and `b` are integers,
  `a + b` is folded
- subtraction `a - b` (left associative): iff both `a` and `b` are integers,
  `a - b` is folded
- multiplication `a * b` (left associative): iff both `a` and `b` are integers,
  `a * b` is folded
- exponentiation / power `a ** b` (right associative): iff both `a` and `b` are
  integers and b is non-negative, `a ** b` is folded

Since (true) division (left associative) in Python always produces a floating
point number, even for `a / 1`, division does not perform symbolic integer
constant-folding in general. However, if for `a / b` both `a` and `b` are
integers and have a greatest common denominator / factor g, i.e. `a = g * c`
and `b = g * d`, the division is symbolically simplified to `c / d`.
Furthermore, `a / 1` is evaluated to `a.0` and `a / -1` to `-a.0`. Since `a.0`
is a floating-point literal, symbolic integer constant folding stops there.

### Numerical evaluation order

Numerical evaluation (after symbolic integer constant folding) evaluates the
quantity of interest literally as written. No clever rewritings are performed.
Therefore, the quantities of interest have strictly defined and reproducible
evaluation rules, which allow writing them to reproduce the same evaluation as
existing (Python) code. For example (where `t1` ... are temporaries):

1. `3.0 + 1 + 7 * (x - 3)`
2. `4.0 + 7 * (x - 3)`
3. `t1 = x - 3.0`
4. `4.0 + 7 * t1`
5. `t2 = 7.0 * t1`
6. `4.0 + t2`

### Numerical functions

Numerical evaluation of the quantities of interest in `compression-safeguards`
is provided by `numpy`.

Some mathematical expressions such as the square root can be written using (a)
exponentiation `x ** 0.5` or (b) the built-in `sqrt(x)` function. It is
preferable to use special built-in functions, where available, since the
safeguards can better understand their meaning and provide better corrections
and higher compression ratios for them.

The operators and functions in the above QoI grammar are evaluated using
`numpy` ufuncs and follow the specification of the `math.h` ISO C standard
(see e.g. <https://pubs.opengroup.org/onlinepubs/9799919799/>):

| QoI function | `numpy` ufunc | `math.h` equivalent | -0.0 behaviour |
| ------------ | ------------- | ------------------- | -------------- |
| `+a` | no-op | |  -0.0 |
| `-a` | `np.negative` | | +0.0 |
| `a + b` | `np.add` | | recessive[^5] |
| `a - b` | `np.subtract` | | recessive[^5] |
| `a * b` | `np.multiply` | | as expected[^6] |
| `a / b` | `np.divide` | | as expected[^7] |
| `a ** b` | `np.power` | `pow` | as expected[^8] |
| `ln` | `np.log` | `log` | -inf |
| `log2` | `np.log2` | `log2` | -inf |
| `log10` | `np.log10` | `log10` | -inf |
| `log(a, base=b)` | `np.divide(np.log(a), np.log(b))` | | as expected |
| `exp` | `np.exp` | `exp` | 1.0 |
| `exp2` | `np.exp2` | `exp2` | 1.0 |
| `exp10` | `np.power(10, a)` | | 1.0 |
| `sqrt` | `np.sqrt` | `sqrt` | -0.0 |
| `square` | `np.square` | | +0.0 |
| `reciprocal` | `np.reciprocal` | | -inf |
| `abs` | `np.abs` | | +0.0 |
| `sign` | `np.sign` | | +0.0 |
| `floor` | `np.floor` | `floor` | -0.0 |
| `ceil` | `np.ceil` | `ceil` | -0.0 |
| `trunc` | `np.trunc` | `trunc` | -0.0 |
| `round_ties_even` | `np.rint` | `rint`[^4] | -0.0 |
| `sin` | `np.sin` | `sin` | -0.0 |
| `cos` | `np.cos` | `cos` | 1.0 |
| `tan` | `np.tan` | `tan` | -0.0 |
| `asin` | `np.arcsin` | `asin` | -0.0 |
| `acos` | `np.arccos` | `acos` | pi/2 |
| `atan` | `np.arctan` | `atan` | -0.0 |
| `sinh` | `np.sinh` | `sinh` | -0.0 |
| `cosh` | `np.cosh` | `cosh` | 1.0 |
| `tanh` | `np.tanh` | `tanh` | -0.0 |
| `asinh` | `np.arcsinh` | `asinh` | -0.0 |
| `acosh` | `np.arccosh` | `acosh` | NaN |
| `atanh` | `np.arctanh` | `atanh` | -0.0 |
| `isfinite` | `np.isfinite` | `isfinite` | 1.0 |
| `isinf` | `np.isinf` | `isinf` | +0.0 |
| `isnan` | `np.isnan` | `isnan` | +0.0 |
| `where` | `np.where` | | as expected |

[^4]: with the round-to-nearest rounding mode
[^5]: -0.0 + -0.0 = -0.0, -0.0 - +0.0 = -0.0, otherwise -0.0 is treated like +0.0
[^6]: a * -0.0 = -a * +0.0
[^7]: a / -0.0 = -a / +0.0, -0.0 / b = +0.0 / -b
[^8]: -0.0 ** (2k+1) = -0.0, -0.0 ** +b = +0.0 otherwise, -0.0 ** (-2k=1) = -inf, -0.0 ** -b = +inf otherwise, a ** -0.0 = 1.0

Furthermore, the array `sum` and `matmul` functions are implemented as explicit
sums over the array elements in natural order, e.g.
`sum([[1, 2], [3, 4]]) = (1 + 2 + 3 + 4)` and
`matmul([[1, 2]], [[3], [4]]) = [[(1 * 3 + 2 * 4)]]`.

### Finite differences

The `finite_difference` function can be used to compute the finite-difference-
approximated derivative over an expression. The finite difference coefficients
for arbitrary orders, accuracies, and grid spacings are derived using the
algorithm from:

> Fornberg, B. (1988). Generation of finite difference formulas on arbitrarily
spaced grids. *Mathematics of Computation*, 51(184), 699-706. Available from:
[doi:10.1090/s0025-5718-1988-0935077-0](https://doi.org/10.1090/s0025-5718-1988-0935077-0).

The computation of the coefficients uses symbolic integer constant folding,
where possible, to produce accurate coefficients on a best-effort basis. Since
the quantities of interest perform no symbolic constant folding on floating
point literals, however, numerical rounding errors can occur when computing the
coefficients for large orders or when using non-integer grid spacings,
custom grid periods, or arbitrary late-bound constant grids. Therefore, the
bitwise exact evaluation of the `finite_difference` function is not yet
specified.
"""

__all__ = [
    "PointwiseQuantityOfInterestExpression",
    "StencilQuantityOfInterestExpression",
]

from typing import NewType

PointwiseQuantityOfInterestExpression = NewType(
    "PointwiseQuantityOfInterestExpression", str
)
"""
Pointwise quantity of interest expression in [`str`][str]ing form, following
the above EBNF grammar for pointwise QoIs.
"""


StencilQuantityOfInterestExpression = NewType(
    "StencilQuantityOfInterestExpression", str
)
"""
Stencil quantity of interest expression in [`str`][str]ing form, following
the above EBNF grammar for stencil QoIs.
"""



"""
Implementations for the provided [`Safeguard`][compression_safeguards.safeguards.abc.Safeguard]s.
"""

__all__ = ["SafeguardKind"]

from enum import Enum

from ..utils.error import TypeCheckError, ctx, lookup_enum_or_raise
from ..utils.typing import JSON
from .abc import Safeguard
from .combinators.all import AllSafeguards
from .combinators.any import AnySafeguard
from .combinators.assume_safe import AssumeAlwaysSafeguard
from .combinators.select import SelectSafeguard
from .pointwise.eb import ErrorBoundSafeguard
from .pointwise.qoi.eb import PointwiseQuantityOfInterestErrorBoundSafeguard
from .pointwise.same import SameValueSafeguard
from .pointwise.sign import SignPreservingSafeguard
from .stencil.qoi.eb import StencilQuantityOfInterestErrorBoundSafeguard


class SafeguardKind(Enum):
    """
    Enumeration of all supported safeguards:
    """

    # same value
    same = SameValueSafeguard
    """Enforce that a special value is exactly preserved."""

    # sign
    sign = SignPreservingSafeguard
    """Enforce that the sign (-1, 0, +1) of each element is preserved."""

    # error bounds
    eb = ErrorBoundSafeguard
    """Enforce a pointwise error bound."""

    qoi_eb_pw = PointwiseQuantityOfInterestErrorBoundSafeguard
    """Enforce an error bound on a pointwise derived quantity of interest."""

    qoi_eb_stencil = StencilQuantityOfInterestErrorBoundSafeguard
    """Enforce an error bound on a derived quantity of interest over a data neighbourhood."""

    # logical combinators
    all = AllSafeguards
    """Enforce that all of the inner safeguards' guarantees are met."""

    any = AnySafeguard
    """Enforce that any one of the inner safeguards' guarantees are met."""

    assume_safe = AssumeAlwaysSafeguard
    """All elements are assumed to always be safe."""

    select = SelectSafeguard
    """Select, pointwise, which safeguard's guarantees to enforce."""

    @staticmethod
    def from_config(config: dict[str, JSON]) -> Safeguard:
        """
        Instantiate a safeguard from a configuration [`dict`][dict].

        The `config` must contain the safeguard's `kind`.

        Parameters
        ----------
        config : dict[str, JSON]
            Configuration of the safeguard.

        Returns
        -------
        safeguard : Safeguard
            Instantiated safeguard.

        Raises
        ------
        ValueError
            if the `config` does not contain the safeguard `kind` or the
            `kind` is unknown.
        TypeCheckError
            if the safeguard `kind` is not a [`str`][str]ing.
        ...
            if instantiating the safeguard raises an exception.
        """

        if "kind" not in config:
            raise ValueError("missing safeguard `kind`") | ctx

        kind = config["kind"]

        with ctx.parameter("kind"):
            TypeCheckError.check_instance_or_raise(kind, str)
            safeguard: type[Safeguard] = lookup_enum_or_raise(
                SafeguardKind,
                kind,  # type: ignore
            ).value

        return safeguard.from_config({p: v for p, v in config.items() if p != "kind"})



"""
Implementation of the [`Safeguards`][compression_safeguards.api.Safeguards], which compute the correction needed to satisfy a set of [`Safeguard`][compression_safeguards.safeguards.abc.Safeguard]s.
"""

__all__ = ["Safeguards"]

import functools
from collections.abc import Collection, Mapping, Set
from typing import Final, Literal, Self, assert_never

import numpy as np
from semver import Version
from typing_extensions import override  # MSPV 3.12

from .safeguards import SafeguardKind
from .safeguards.abc import Safeguard
from .safeguards.pointwise.abc import PointwiseSafeguard
from .safeguards.stencil import BoundaryCondition, NeighbourhoodAxis
from .safeguards.stencil.abc import StencilSafeguard
from .utils._compat import _ones, _zeros
from .utils.bindings import Bindings, Parameter, Value
from .utils.cast import as_bits
from .utils.error import (
    LateBoundParameterResolutionError,
    SafeguardsSafetyBug,
    TypeSetError,
    ctx,
)
from .utils.intervals import IntervalUnion  # noqa: TC001
from .utils.typing import JSON, C, S, T


class Safeguards:
    """
    Collection of [`Safeguard`][compression_safeguards.safeguards.abc.Safeguard]s.

    Parameters
    ----------
    safeguards : Collection[dict[str, JSON] | Safeguard]
        The safeguards that will be applied. They can either be passed as a
        safeguard configuration [`dict`][dict] or an already initialized
        [`Safeguard`][...safeguards.abc.Safeguard].

        Please refer to the [`SafeguardKind`][...safeguards.SafeguardKind] for
        an enumeration of all supported safeguards.
    _version : ...
        The version of the safeguards. Do not provide this parameter explicitly.

    Raises
    ------
    ValueError
        if the safeguards are instantiated with a configuration from an
        incompatible version of the safeguards.
    NotImplementedError
        if the `safeguards` contain an unsupported kind of safeguard
        (currently: pointwise and stencil safeguards are supported).
    ...
        if instantiating a safeguard raises an exception.
    """

    __slots__: tuple[str, ...] = ("_pointwise_safeguards", "_stencil_safeguards")
    _pointwise_safeguards: tuple[PointwiseSafeguard, ...]
    _stencil_safeguards: tuple[StencilSafeguard, ...]

    def __init__(
        self,
        *,
        safeguards: Collection[dict[str, JSON] | Safeguard],
        _version: None | str | Version = None,
    ) -> None:
        if _version is not None:
            _version = (
                _version if isinstance(_version, Version) else Version.parse(_version)
            )
            if not _version.is_compatible(self.version):
                raise (
                    ValueError(
                        f"{_version} is not semantic-versioning-compatible with "
                        + f"the safeguards version {self.version}"
                    )
                    | ctx
                )

        safeguards_: list[Safeguard] = []
        with ctx.parameter("safeguards"):
            for i, safeguard in enumerate(safeguards):
                with ctx.index(i):
                    safeguards_.append(
                        safeguard
                        if isinstance(safeguard, Safeguard)
                        else SafeguardKind.from_config(safeguard)
                    )

        self._pointwise_safeguards = tuple(
            safeguard
            for safeguard in safeguards_
            if isinstance(safeguard, PointwiseSafeguard)
        )
        self._stencil_safeguards = tuple(
            safeguard
            for safeguard in safeguards_
            if isinstance(safeguard, StencilSafeguard)
        )
        unsupported_safeguards = [
            safeguard
            for safeguard in safeguards_
            if not isinstance(safeguard, PointwiseSafeguard | StencilSafeguard)
        ]

        if len(unsupported_safeguards) > 0:
            with ctx.parameter("safeguards"):
                raise NotImplementedError(unsupported_safeguards) | ctx

    @property
    def safeguards(self) -> Collection[Safeguard]:
        """
        The collection of safeguards.
        """

        return self._pointwise_safeguards + self._stencil_safeguards

    @property
    def late_bound(self) -> Set[Parameter]:
        """
        The set of late-bound parameters that the safeguards have.

        Late-bound parameters are only bound when computing the correction, in
        contrast to the normal early-bound parameters that are configured
        during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        return frozenset(b for s in self.safeguards for b in s.late_bound)

    @property
    def builtin_late_bound(self) -> Set[Parameter]:
        """
        The set of built-in late-bound constants that the safeguards provide
        automatically, which include `$x` and `$X`.
        """

        return frozenset([Parameter("$x"), Parameter("$X")])

    @property
    def version(self) -> Version:
        """
        The semantic version [^1] of the safeguards provided by this package,
        which covers

        - the guarantees provided by the safeguards (can only be weakened in a
          new breaking major release)
        - the configurations of the safeguards (can be extended backwards-
          compatibly in a new minor release)
        - the format of the safeguards corrections (can only be changed in a
          new breaking major release)

        The [`Safeguards`][..] can only load configurations for and apply
        corrections produced by safeguards with a compatible semantic version.

        Note that the version of the safeguards may be different from the
        version of the `compression-safeguards` package, which may make changes
        to the implementation or programmatic API without needing to increase
        the version of the safeguards.

        [^1]: <https://semver.org>
        """

        return _SAFEGUARDS_VERSION

    @staticmethod
    def supported_dtypes() -> frozenset[np.dtype[np.number]]:
        """
        The set of numpy [`dtype`][numpy.dtype]s that the safeguards support.
        """

        return _SUPPORTED_DTYPES

    def correction_dtype_for_data(self, dtype: np.dtype[T]) -> np.dtype[C]:
        """
        Compute the dtype of the correction for data of the provided `dtype`.

        The correction dtype is the corresponding binary unsigned integer data
        type with the same bit size.

        Parameters
        ----------
        dtype : np.dtype[T]
            The dtype of the data.

        Returns
        -------
        correction : np.dtype[C]
            The dtype of the correction.
        """

        return as_bits(np.array((), dtype=dtype)).dtype

    def check(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Mapping[str | Parameter, Value] | Bindings = Bindings.EMPTY,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> bool:
        """
        Check if the `prediction` array upholds the properties enforced by the safeguards with respect to the `data` array.

        The `data` array must contain the complete data, i.e. not just a chunk
        of data, so that non-pointwise safeguards are correctly applied. Please
        use the [`check_chunk`][..check_chunk] method instead when working with
        individual chunks of data.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            The data array, relative to which the safeguards are checked.
        prediction : np.ndarray[S, np.dtype[T]]
            The prediction array for which the safeguards are checked.
        late_bound : Mapping[str | Parameter, Value] | Bindings
            The bindings for all late-bound parameters of the safeguards.

            The bindings must resolve all late-bound parameters and include no
            extraneous parameters.

            The safeguards automatically provide the `$x` and `$X` built-in
            constants, which must not be included.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : bool
            `True` if the check succeeded.

        Raises
        ------
        TypeSetError
            if the `data` uses an unsupported data type.
        ValueError
            if the `data`'s dtype or shape do not match the `prediction`'s.
        RuntimeError
            if the `data` array is chunked.
        LateBoundParameterResolutionError
            if `late_bound` does not resolve all late-bound parameters of the
            safeguards or includes any extraneous parameters.
        ...
            if checking a safeguard raises an exception.
        """

        late_bound = self._prepare_non_chunked_bindings(
            data=data,
            prediction=prediction,
            late_bound=late_bound,
            description="checking the safeguards",
            chunked_method_name="check_chunk",
        )

        for safeguard in self.safeguards:
            if not safeguard.check(
                data, prediction, late_bound=late_bound, where=where
            ):
                return False

        return True

    def compute_correction(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Mapping[str | Parameter, Value] | Bindings = Bindings.EMPTY,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[C]]:
        """
        Compute the correction required to make the `prediction` array satisfy the safeguards relative to the `data` array.

        The `data` array must contain the complete data, i.e. not just a chunk
        of data, so that non-pointwise safeguards are correctly applied. Please
        use the [`compute_chunked_correction`][..compute_chunked_correction]
        method instead when working with individual chunks of data.

        The correction is defined as `as_bits(corrected) - as_bits(prediction)`
        using wrapping unsigned integer arithmetic. It has the corresponding
        binary unsigned integer data type with the same bit size. The
        correction can be applied using [`apply_correction`][..apply_correction]
        to get the corrected `prediction`.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            The data array, relative to which the safeguards are enforced.
        prediction : np.ndarray[S, np.dtype[T]]
            The prediction array for which the correction is computed.
        late_bound : Mapping[str | Parameter, Value] | Bindings
            The bindings for all late-bound parameters of the safeguards.

            The bindings must resolve all late-bound parameters and include no
            extraneous parameters.

            The safeguards automatically provide the `$x` and `$X` built-in
            constants, which must not be included.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the correction at data points where the condition
            is [`True`][True].

        Returns
        -------
        correction : np.ndarray[S, np.dtype[C]]
            The correction array.

        Raises
        ------
        TypeSetError
            if the `data` uses an unsupported data type.
        ValueError
            if the `data`'s dtype or shape do not match the `prediction`'s.
        RuntimeError
            if the `data` array is chunked.
        LateBoundParameterResolutionError
            if `late_bound` does not resolve all late-bound parameters of the
            safeguards or includes any extraneous parameters.
        ...
            if computing the correction for a safeguard raises an exception.
        """
        # explicitly do not document that SafeguardsSafetyBug can be raised

        late_bound = self._prepare_non_chunked_bindings(
            data=data,
            prediction=prediction,
            late_bound=late_bound,
            description="computing the safeguards correction",
            chunked_method_name="compute_chunked_correction",
        )

        all_ok = True
        for safeguard in self.safeguards:
            if not safeguard.check(
                data, prediction, late_bound=late_bound, where=where
            ):
                all_ok = False
                break

        if all_ok:
            return _zeros(data.shape, self.correction_dtype_for_data(data.dtype))

        # ensure we don't accidentally forget to handle new kinds of safeguards here
        assert len(self.safeguards) == len(self._pointwise_safeguards) + len(
            self._stencil_safeguards
        )

        all_intervals: list[IntervalUnion[T, int, int]] = []
        for safeguard in self._pointwise_safeguards + self._stencil_safeguards:
            intervals = safeguard.compute_safe_intervals(
                data, late_bound=late_bound, where=where
            )
            if not np.all(intervals.contains(data)):
                raise (
                    SafeguardsSafetyBug(
                        f"the safe intervals for the {safeguard!r} safeguard "
                        + "do not contain the original data"
                    )
                    | ctx
                )
            all_intervals.append(intervals)

        combined_intervals = all_intervals[0]
        for intervals in all_intervals[1:]:
            combined_intervals = combined_intervals.intersect(intervals)
        corrected = combined_intervals.pick(prediction)

        for safeguard, intervals in zip(self.safeguards, all_intervals):
            if not np.all(intervals.contains(corrected)):
                raise (
                    SafeguardsSafetyBug(
                        f"the safe intervals for the {safeguard!r} safeguard "
                        + "do not contain the corrected array"
                    )
                    | ctx
                )
            if not safeguard.check(data, corrected, late_bound=late_bound, where=where):
                raise (
                    SafeguardsSafetyBug(
                        f"the check for the {safeguard!r} safeguard fails "
                        + "with the corrected array"
                    )
                    | ctx
                )

        prediction_bits: np.ndarray[S, np.dtype[C]] = as_bits(prediction)
        corrected_bits: np.ndarray[S, np.dtype[C]] = as_bits(corrected)

        with np.errstate(
            over="ignore",
            under="ignore",
        ):
            return corrected_bits - prediction_bits

    def _prepare_non_chunked_bindings(
        self,
        *,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        late_bound: Mapping[str | Parameter, Value] | Bindings,
        description: str,
        chunked_method_name: str,
    ) -> Bindings:
        TypeSetError.check_dtype_or_raise(data.dtype, _SUPPORTED_DTYPES)

        # ensure we don't accidentally forget to handle new kinds of safeguards here
        assert len(self.safeguards) == len(self._pointwise_safeguards) + len(
            self._stencil_safeguards
        )

        with ctx.parameter("prediction"):
            if prediction.dtype != data.dtype:
                raise ValueError("prediction.dtype must match data.dtype") | ctx
            if prediction.shape != data.shape:
                raise ValueError("prediction.shape must match data.shape") | ctx

        if len(self._stencil_safeguards) > 0 and getattr(data, "chunked", False):
            with ctx.parameter("data"):
                raise (
                    RuntimeError(
                        f"{description} for an individual chunk in a chunked "
                        + "array is unsafe when using stencil safeguards since "
                        + "their safety requirements cannot be guaranteed "
                        + f"across chunk boundaries; use {chunked_method_name} "
                        + "instead"
                    )
                    | ctx
                )

        late_bound = (
            late_bound if isinstance(late_bound, Bindings) else Bindings(**late_bound)
        )

        late_bound_reqs = self.late_bound
        late_bound_builtin = {
            p: data for p in late_bound_reqs if p in self.builtin_late_bound
        }
        late_bound_reqs = frozenset(late_bound_reqs - late_bound_builtin.keys())
        late_bound_keys = frozenset(late_bound.parameters())
        LateBoundParameterResolutionError.check_or_raise(
            late_bound_reqs, late_bound_keys
        )

        if len(late_bound_builtin) > 0:
            late_bound = late_bound.update(
                **{str(p): v for p, v in late_bound_builtin.items()}
            )

        return late_bound

    def apply_correction(
        self,
        prediction: np.ndarray[S, np.dtype[T]],
        correction: np.ndarray[S, np.dtype[C]],
    ) -> np.ndarray[S, np.dtype[T]]:
        """
        Apply the `correction` to the `prediction` to satisfy the safeguards for which the `correction` was computed.

        The `prediction` must be bitwise equivalent to the `prediction` that
        was used to compute the `correction`.

        This method is guaranteed to work for chunked data as well, i.e.
        applying a chunk of the `correction` to the corresponding chunk of the
        `prediction` produces the correct result.

        The `correction` is applied with
        `from_bits(as_bits(prediction) + correction)` using wrapping unsigned
        integer arithmetic.

        Parameters
        ----------
        prediction : np.ndarray[S, np.dtype[T]]
            The prediction array for which the correction has been computed.
        correction : np.ndarray[S, np.dtype[C]]
            The correction array.

        Returns
        -------
        corrected : np.ndarray[S, np.dtype[T]]
            The corrected array, which satisfies the safeguards.

        Raises
        ------
        ValueError
            if the `correction`'s shape dos not match the `prediction`'s, or if
            the `correction`'s dtype does not match the correction dtype for
            the `prediction`'s dtype.
        """

        with ctx.parameter("correction"):
            if correction.dtype != self.correction_dtype_for_data(prediction.dtype):
                raise (
                    ValueError(
                        "correction.dtype must match the correction dtype for "
                        + "prediction.dtype"
                    )
                    | ctx
                )
            if correction.shape != prediction.shape:
                raise ValueError("correction.shape must match prediction.shape") | ctx

        prediction_bits: np.ndarray[S, np.dtype[C]] = as_bits(prediction)
        correction_bits = correction

        with np.errstate(
            over="ignore",
            under="ignore",
        ):
            corrected_bits: np.ndarray[S, np.dtype[C]] = np.add(
                prediction_bits, correction_bits
            )

        return corrected_bits.view(prediction.dtype)

    @functools.lru_cache
    def compute_required_stencil_for_chunked_correction(
        self, data_shape: tuple[int, ...]
    ) -> tuple[
        tuple[
            Literal[BoundaryCondition.valid, BoundaryCondition.wrap], NeighbourhoodAxis
        ],
        ...,
    ]:
        """
        Compute the shape of the stencil neighbourhood around chunks of the complete data that is required to compute the chunked corrections.

        For each data dimension, the stencil might require either a
        [valid][....safeguards.stencil.BoundaryCondition.valid] or
        [wrapping][....safeguards.stencil.BoundaryCondition.wrap] boundary
        condition.

        This method also checks that the data shape is compatible with the
        safeguards.

        Parameters
        ----------
        data_shape : tuple[int, ...]
            The shape of the complete data.

        Returns
        -------
        stencil_shape : tuple[tuple[Literal[BoundaryCondition.valid, BoundaryCondition.wrap], NeighbourhoodAxis], ...]
            The shape of the required stencil neighbourhood around each chunk.

        Raises
        ------
        ...
            if computing the stencil neighbourhood for a safeguard raises an
            exception.
        """

        neighbourhood: list[
            tuple[
                Literal[BoundaryCondition.valid, BoundaryCondition.wrap],
                NeighbourhoodAxis,
            ]
        ] = [(BoundaryCondition.valid, NeighbourhoodAxis(0, 0)) for _ in data_shape]

        # ensure we don't accidentally forget to handle new kinds of safeguards here
        assert len(self.safeguards) == len(self._pointwise_safeguards) + len(
            self._stencil_safeguards
        )

        # pointwise safeguards don't impose any stencil neighbourhood
        #  requirements

        for safeguard in self._stencil_safeguards:
            for i, bs in enumerate(
                safeguard.compute_check_neighbourhood_for_data_shape(data_shape)
            ):
                for b, s in bs.items():
                    n_before, n_after = (
                        neighbourhood[i][1].before,
                        neighbourhood[i][1].after,
                    )

                    # we now know that the safeguards have a stencil of
                    #  [before, ..., x, ..., after]
                    # this stencil is sufficient to compute the safeguards for x
                    #
                    # BUT the elements in before and after can also back-
                    #  contribute to the safe intervals of x,
                    # so we need to ensure that all elements in the stencil can
                    #  also apply the safeguards, i.e. they also need their
                    #  stencil supplied
                    # in particular, we need to ensure that the elements before
                    #  x, which will look after elements to the right, get
                    #  their stencil, and that the elements after x, which will
                    #  look before elements to the left, get their stencil too,
                    #  since they will back-contribute their intervals to x
                    #
                    # therefore we actually need to ~double~ the stencil to
                    # [after+before, ..., before, ..., x, ..., after, ..., before+after]

                    match b:
                        case (
                            BoundaryCondition.valid
                            | BoundaryCondition.constant
                            | BoundaryCondition.edge
                        ):
                            # nothing special, but we do need to extend the stencil
                            neighbourhood[i] = (
                                neighbourhood[i][0],
                                NeighbourhoodAxis(
                                    before=max(n_before, s.after + s.before),
                                    after=max(n_after, s.before + s.after),
                                ),
                            )
                        case BoundaryCondition.reflect:
                            # reflect:           [ 1, 2, 3 ]
                            #       -> [..., 3, 2, 1, 2, 3, 2, 1, ...]
                            # worst case, the reflection on the left exits the
                            #  chunk on the right, and same for on the right
                            # so we need to extend with max(before, after) at
                            #  both ends
                            # if we have before=1 and after=0 and focus on x=3,
                            #  this would give us the stencil 2, 3|
                            # however, when the reflect boundary is applied, it
                            #  would be 3, 2, 3| and 2 would now look at 3 and
                            #  back-contribute to 3 even though it should not
                            # if we have before=2 and after=0 and focus on x=3
                            #  in -1, 0, 1, 2, 3 then 1, 2, 3| would become
                            #  3, 2, 1, 2, 3| which is incorrect; 0, 1, 2, 3|
                            #  would become 2, 1, 0, 1, 2, 3 which is also
                            #  incorrect but good enough to no longer back-
                            #  contribute
                            # therefore, we need to extend by at least one
                            #  beyond the stencil if the stencil is non-zero
                            # but since we need before+after on both sides, the
                            #  max cancels out (and we also cannot optimise
                            #  based on the size of the chunk)
                            neighbourhood[i] = (
                                neighbourhood[i][0],
                                NeighbourhoodAxis(
                                    before=max(
                                        n_before,
                                        max(s.after, bool(s.before)) + s.before,
                                    ),
                                    after=max(
                                        n_after, max(s.before, bool(s.after)) + s.after
                                    ),
                                ),
                            )
                        case BoundaryCondition.symmetric:
                            # symmetric:         [ 1, 2, 3 ]
                            #       -> [..., 2, 1, 1, 2, 3, 3, 2, ...]
                            # similar to reflect, but the edge is repeated,
                            #  meaning that accidental back-contribution is not
                            #  possible
                            # worst case, the symmetry on the left exits the
                            #  chunk on the right, and same for on the right
                            # but since we need before+after on both sides, the
                            #  max cancels out (and we also cannot optimise
                            #  based on the size of the chunk)
                            neighbourhood[i] = (
                                neighbourhood[i][0],
                                NeighbourhoodAxis(
                                    before=max(n_before, s.after + s.before),
                                    after=max(n_after, s.before + s.after),
                                ),
                            )
                        case BoundaryCondition.wrap:
                            # we need to extend the stencil and tell xarray and
                            #  remember that we will need a wrapping / periodic
                            #  boundary
                            neighbourhood[i] = (
                                # only require the wrapping boundary condition
                                #  if there actually is a boundary
                                BoundaryCondition.wrap
                                if s.before > 0 or s.after > 0
                                else neighbourhood[i][0],
                                NeighbourhoodAxis(
                                    before=max(n_before, s.after + s.before),
                                    after=max(n_after, s.before + s.after),
                                ),
                            )
                        case _:
                            assert_never(b)

        return tuple(neighbourhood)

    def check_chunk(
        self,
        data_chunk: np.ndarray[S, np.dtype[T]],
        prediction_chunk: np.ndarray[S, np.dtype[T]],
        *,
        data_shape: tuple[int, ...],
        chunk_offset: tuple[int, ...],
        chunk_stencil: tuple[
            tuple[
                Literal[BoundaryCondition.valid, BoundaryCondition.wrap],
                NeighbourhoodAxis,
            ],
            ...,
        ],
        late_bound_chunk: Mapping[str | Parameter, Value] | Bindings = Bindings.EMPTY,
        where_chunk: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> bool:
        """
        Check if the `prediction_chunk` array chunk upholds the properties enforced by the safeguards with respect to the `data_chunk` array chunk.

        Both the `data_chunk` and `prediction_chunk` contain the stencil around
        the chunk, and the shape of this applied stencil is specified in the
        `chunk_stencil` parameter. This stencil must be compatible with the
        required stencil returned by
        [`compute_required_stencil_for_chunked_correction(data_shape)`][..compute_required_stencil_for_chunked_correction]:
        - a wrapping boundary is always compatible with a valid boundary.
        - a larger stencil is always compatible with a smaller stencil.
        - a smaller stencil is sometimes compatible with a larger stencil, iff
          the smaller stencil is near the entire data boundary and still
          includes all required elements; for instance, providing the entire
          data as a single chunk with no stencil is always compatible with any
          stencil

        This advanced method should only be used when working with individual
        chunks of data, for non-chunked data please use the simpler and more
        efficient [`check`][..check] method instead.

        Parameters
        ----------
        data_chunk : np.ndarray[S, np.dtype[T]]
            A stencil-extended chunk from the data array, relative to which the
            safeguards are checked.
        prediction_chunk : np.ndarray[S, np.dtype[T]]
            The corresponding stencil-extended chunk from the prediction array
            for which the safeguards are checked.
        data_shape : tuple[int, ...]
            The shape of the entire data array, i.e. not just the chunk,
            without any stencil.
        chunk_offset : tuple[int, ...]
            The offset of the non-stencil-extended chunk inside the entire
            array. For arrays going from left to right, bottom to top, ..., the
            offset is the index of the bottom left element in the entire array.
        chunk_stencil : tuple[tuple[Literal[BoundaryCondition.valid, BoundaryCondition.wrap], NeighbourhoodAxis], ...]
            The shape of the stencil neighbourhood that was applied around the
            chunk in `data_chunk` and `prediction_chunk`.
        late_bound_chunk : Mapping[str | Parameter, Value] | Bindings
            The bindings for all late-bound parameters of the safeguards.

            The bindings must resolve all late-bound parameters and include no
            extraneous parameters.

            If a binding resolves to an array, it must be the corresponding
            chunk of the entire late-bound array.

            The safeguards automatically provide the `$x` and `$X` built-in
            constants, which must not be included.
        where_chunk : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        chunk_ok : bool
            `True` if the check succeeded for the chunk.

        Raises
        ------
        TypeSetError
            if the `data_chunk` uses an unsupported data type.
        ValueError
            if the `data_chunk`'s dtype or shape do not match the
            `prediction_chunk`'s, or if the length of the `data_shape`,
            `chunk_offset`, or `chunk_stencil` do not match the dimensionality
            of the `data`.
        ValueError
            if the `chunk_stencil` is not compatible with the required stencil.
        LateBoundParameterResolutionError
            if `late_bound_chunk` does not resolve all late-bound parameters of
            the safeguards or includes any extraneous parameters.
        ValueError
            if any `late_bound_chunk` array could not be broadcast to the
            `data_chunk`'s shape.
        ...
            if checking a safeguard raises an exception.
        """

        (
            data_chunk_,
            prediction_chunk_,
            late_bound_chunk,
            where_chunk_,
            non_stencil_indices,
        ) = self._prepare_stencil_chunked_arrays_and_bindings(
            data_chunk=data_chunk,
            prediction_chunk=prediction_chunk,
            data_shape=data_shape,
            chunk_offset=chunk_offset,
            chunk_stencil=chunk_stencil,
            late_bound_chunk=late_bound_chunk,
            where_chunk=where_chunk,
        )

        # ensure we don't accidentally forget to handle new kinds of safeguards here
        assert len(self.safeguards) == len(self._pointwise_safeguards) + len(
            self._stencil_safeguards
        )

        all_ok = _ones(
            data_chunk_[tuple(non_stencil_indices)].shape, dtype=np.dtype(np.bool)
        )

        # we need to use pointwise checks here so that we can only look at the
        #  non-stencil check results
        for safeguard in self._pointwise_safeguards + self._stencil_safeguards:
            all_ok &= safeguard.check_pointwise(
                data_chunk_,
                prediction_chunk_,
                late_bound=late_bound_chunk,
                where=where_chunk_,
            )[tuple(non_stencil_indices)]

            if not np.all(all_ok):
                return False

        return True

    def compute_chunked_correction(
        self,
        data_chunk: np.ndarray[S, np.dtype[T]],
        prediction_chunk: np.ndarray[S, np.dtype[T]],
        *,
        data_shape: tuple[int, ...],
        chunk_offset: tuple[int, ...],
        chunk_stencil: tuple[
            tuple[
                Literal[BoundaryCondition.valid, BoundaryCondition.wrap],
                NeighbourhoodAxis,
            ],
            ...,
        ],
        any_chunk_check_failed: bool,
        late_bound_chunk: Mapping[str | Parameter, Value] | Bindings = Bindings.EMPTY,
        where_chunk: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[tuple[int, ...], np.dtype[C]]:
        """
        Compute the correction required to make the `prediction_chunk` array chunk satisfy the safeguards relative to the `data_chunk` array chunk.

        Both the `data_chunk` and `prediction_chunk` contain the stencil around
        the chunk, and the shape of this applied stencil is specified in the
        `chunk_stencil` parameter. This stencil must be compatible with the
        required stencil returned by
        [`compute_required_stencil_for_chunked_correction(data_shape)`][..compute_required_stencil_for_chunked_correction]:
        - a wrapping boundary is always compatible with a valid boundary.
        - a larger stencil is always compatible with a smaller stencil.
        - a smaller stencil is sometimes compatible with a larger stencil, iff
          the smaller stencil is near the entire data boundary and still
          includes all required elements; for instance, providing the entire
          data as a single chunk with no stencil is always compatible with any
          stencil

        This advanced method should only be used when working with individual
        chunks of data, for non-chunked data please use the simpler and more
        efficient [`compute_correction`][..compute_correction] method instead.

        The (chunked) correction is defined as
        `as_bits(corrected) - as_bits(prediction)` using wrapping unsigned
        integer arithmetic. It has the corresponding binary unsigned integer
        data type with the same bit size. The chunked correction can be applied
        using [`apply_correction`][..apply_correction] to get the corrected
        chunked `prediction`.

        Parameters
        ----------
        data_chunk : np.ndarray[S, np.dtype[T]]
            A stencil-extended chunk from the data array, relative to which the
            safeguards are enforced.
        prediction_chunk : np.ndarray[S, np.dtype[T]]
            The corresponding stencil-extended chunk from the prediction array
            for which the correction is computed.
        data_shape : tuple[int, ...]
            The shape of the entire data array, i.e. not just the chunk.
        chunk_offset : tuple[int, ...]
            The offset of the non-stencil-extended chunk inside the entire
            array. For arrays going from left to right, bottom to top, ..., the
            offset is the index of the bottom left element in the entire array.
        chunk_stencil : tuple[tuple[Literal[BoundaryCondition.valid, BoundaryCondition.wrap], NeighbourhoodAxis], ...]
            The shape of the stencil neighbourhood that was applied around the
            chunk in `data_chunk` and `prediction_chunk`.
        late_bound_chunk : Mapping[str | Parameter, Value] | Bindings
            The bindings for all late-bound parameters of the safeguards.

            The bindings must resolve all late-bound parameters and include no
            extraneous parameters.

            If a binding resolves to an array, it must be the corresponding
            chunk of the entire late-bound array.

            The safeguards automatically provide the `$x` and `$X` built-in
            constants, which must not be included.
        where_chunk : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the correction at data points where the condition is
            [`True`][True].

        Returns
        -------
        correction_chunk : np.ndarray[tuple[int, ...], np.dtype[C]]
            The correction array chunk. The correction chunk is truncated to
            remove the stencil, i.e. it only contains the correction for the
            non-stencil-extended chunk.

        Raises
        ------
        TypeSetError
            if the `data_chunk` uses an unsupported data type.
        ValueError
            if the `data_chunk`'s dtype or shape do not match the
            `prediction_chunk`'s, or if the length of the `data_shape`,
            `chunk_offset`, or `chunk_stencil` do not match the dimensionality
            of the `data`.
        ValueError
            if the `chunk_stencil` is not compatible with the required stencil.
        LateBoundParameterResolutionError
            if `late_bound_chunk` does not resolve all late-bound parameters of
            the safeguards or includes any extraneous parameters.
        ValueError
            if any `late_bound_chunk` array could not be broadcast to the
            `data_chunk`'s shape.
        ...
            if computing the correction for a safeguard raises an exception.
        """
        # explicitly do not document that SafeguardsSafetyBug can be raised

        (
            data_chunk_,
            prediction_chunk_,
            late_bound_chunk,
            where_chunk_,
            non_stencil_indices,
        ) = self._prepare_stencil_chunked_arrays_and_bindings(
            data_chunk=data_chunk,
            prediction_chunk=prediction_chunk,
            data_shape=data_shape,
            chunk_offset=chunk_offset,
            chunk_stencil=chunk_stencil,
            late_bound_chunk=late_bound_chunk,
            where_chunk=where_chunk,
        )

        # if no chunk requires a correction, this one doesn't either
        if not any_chunk_check_failed:
            return _zeros(
                data_chunk_[tuple(non_stencil_indices)].shape,
                self.correction_dtype_for_data(data_chunk_.dtype),
            )

        safeguard: Safeguard

        # if only pointwise safeguards are used, check if we need to correct
        #  this chunk
        if len(self._pointwise_safeguards) == len(self.safeguards):
            all_ok = _ones(
                data_chunk_[tuple(non_stencil_indices)].shape, dtype=np.dtype(np.bool)
            )

            # we need to use pointwise checks here so that we can only look at the
            #  non-stencil check results
            for safeguard in self._pointwise_safeguards:
                all_ok &= safeguard.check_pointwise(
                    data_chunk_,
                    prediction_chunk_,
                    late_bound=late_bound_chunk,
                    where=where_chunk_,
                )[tuple(non_stencil_indices)]

                if not np.all(all_ok):
                    break

            if np.all(all_ok):
                return _zeros(
                    data_chunk_[tuple(non_stencil_indices)].shape,
                    self.correction_dtype_for_data(data_chunk_.dtype),
                )

        # otherwise, correct the chunk
        # if stencil safeguards are used, then any chunk needing a correction
        #  requires all chunks to be corrected

        # ensure we don't accidentally forget to handle new kinds of safeguards here
        assert len(self.safeguards) == len(self._pointwise_safeguards) + len(
            self._stencil_safeguards
        )

        all_intervals: list[IntervalUnion[T, int, int]] = []
        for safeguard in self._pointwise_safeguards + self._stencil_safeguards:
            intervals = safeguard.compute_safe_intervals(
                data_chunk_, late_bound=late_bound_chunk, where=where_chunk_
            )
            if not np.all(intervals.contains(data_chunk_)):
                raise (
                    SafeguardsSafetyBug(
                        f"the safe intervals for the {safeguard!r} safeguard "
                        + "must contain the original data chunk"
                    )
                    | ctx
                )
            all_intervals.append(intervals)

        combined_intervals = all_intervals[0]
        for intervals in all_intervals[1:]:
            combined_intervals = combined_intervals.intersect(intervals)
        corrected_chunk = combined_intervals.pick(prediction_chunk_)

        for safeguard, intervals in zip(self.safeguards, all_intervals):
            if not np.all(intervals.contains(corrected_chunk)):
                raise (
                    SafeguardsSafetyBug(
                        f"the safe intervals for the {safeguard!r} safeguard "
                        + "do not contain the corrected array chunk"
                    )
                    | ctx
                )
            if not safeguard.check(
                data_chunk_,
                corrected_chunk,
                late_bound=late_bound_chunk,
                where=where_chunk_,
            ):
                raise (
                    SafeguardsSafetyBug(
                        f"the check for the {safeguard!r} safeguard fails "
                        + "with the corrected array chunk"
                    )
                    | ctx
                )

        prediction_chunk_bits: np.ndarray[tuple[int, ...], np.dtype[C]] = as_bits(
            prediction_chunk_
        )
        corrected_chunk_bits: np.ndarray[tuple[int, ...], np.dtype[C]] = as_bits(
            corrected_chunk
        )

        with np.errstate(
            over="ignore",
            under="ignore",
        ):
            return (corrected_chunk_bits - prediction_chunk_bits)[
                tuple(non_stencil_indices)
            ]

    def _prepare_stencil_chunked_arrays_and_bindings(
        self,
        *,
        data_chunk: np.ndarray[S, np.dtype[T]],
        prediction_chunk: np.ndarray[S, np.dtype[T]],
        data_shape: tuple[int, ...],
        chunk_offset: tuple[int, ...],
        chunk_stencil: tuple[
            tuple[
                Literal[BoundaryCondition.valid, BoundaryCondition.wrap],
                NeighbourhoodAxis,
            ],
            ...,
        ],
        late_bound_chunk: Mapping[str | Parameter, Value] | Bindings,
        where_chunk: Literal[True] | np.ndarray[S, np.dtype[np.bool]],
    ) -> tuple[
        np.ndarray[tuple[int, ...], np.dtype[T]],
        np.ndarray[tuple[int, ...], np.dtype[T]],
        Bindings,
        Literal[True] | np.ndarray[tuple[int, ...], np.dtype[np.bool]],
        tuple[slice, ...],
    ]:
        TypeSetError.check_dtype_or_raise(data_chunk.dtype, _SUPPORTED_DTYPES)

        with ctx.parameter("prediction_chunk"):
            if prediction_chunk.dtype != data_chunk.dtype:
                raise (
                    ValueError("prediction_chunk.dtype must match data_chunk.dtype")
                    | ctx
                )
            if prediction_chunk.shape != data_chunk.shape:
                raise (
                    ValueError("prediction_chunk.shape must match data_chunk.shape")
                    | ctx
                )
        with ctx.parameter("data_shape"):
            if len(data_shape) != data_chunk.ndim:
                raise ValueError("len(data_shape) must match data_chunk.ndim") | ctx
        with ctx.parameter("chunk_offset"):
            if len(chunk_offset) != data_chunk.ndim:
                raise ValueError("len(chunk_offset) must match data_chunk.ndim") | ctx
        with ctx.parameter("chunk_stencil"):
            if len(chunk_stencil) != data_chunk.ndim:
                raise ValueError("len(chunk_stencil) must match data_chunk.ndim") | ctx

        chunk_shape: tuple[int, ...] = tuple(
            a - s[1].before - s[1].after
            for a, s in zip(data_chunk.shape, chunk_stencil)
        )

        required_stencil = self.compute_required_stencil_for_chunked_correction(
            data_shape
        )

        stencil_indices: list[slice] = []
        stencil_roll: list[int] = []
        non_stencil_indices: list[slice] = []

        # (1): check that the chunk stencil is compatible with the required
        #       stencil
        #      this is not trivial since we need to account for huge chunks
        #       where downgrading the stencil can work out
        # (2): compute indices to extract just the needed data and data+stencil
        with ctx.parameter("chunk_stencil"):
            for i, (c, r) in enumerate(zip(chunk_stencil, required_stencil)):
                with ctx.index(i):
                    # complete chunks that span the entire data along the axis
                    #  are always allowed
                    if (
                        c[0] in (BoundaryCondition.valid, BoundaryCondition.wrap)
                        and c[1].before == 0
                        and c[1].after == 0
                        and chunk_shape[i] == data_shape[i]
                    ):
                        stencil_indices.append(slice(None))
                        stencil_roll.append(0)
                        non_stencil_indices.append(slice(None))
                        continue

                    match c[0]:
                        case BoundaryCondition.valid:
                            # we need to check that we requested a valid
                            #  boundary, which is only compatible with itself
                            # and that the stencil is large enough
                            if r[0] != BoundaryCondition.valid:
                                raise (
                                    ValueError(
                                        f"requires {r[0].name} stencil "
                                        + "boundary but found valid boundary"
                                    )
                                    | ctx
                                )

                            # what is the required stencil after adjusting for
                            #  near-boundary stencil truncation?
                            rs = NeighbourhoodAxis(
                                before=min(chunk_offset[i], r[1].before),
                                after=min(
                                    r[1].after,
                                    data_shape[i] - chunk_shape[i] - chunk_offset[i],
                                ),
                            )
                            if c[1].before < rs.before:
                                with ctx.parameter("before"):
                                    raise (
                                        ValueError(
                                            "requires stencil with at least "
                                            + f"{rs.before} point(s) before "
                                            + f"but received only {c[1].before}"
                                        )
                                        | ctx
                                    )
                            if c[1].after < rs.after:
                                with ctx.parameter("after"):
                                    raise (
                                        ValueError(
                                            "requires stencil with at least "
                                            + f"{rs.after} point(s) after but "
                                            + f"received only {c[1].after}"
                                        )
                                        | ctx
                                    )

                            stencil_indices.append(
                                slice(
                                    c[1].before - rs.before,
                                    None
                                    if c[1].after == rs.after
                                    else rs.after - c[1].after,
                                )
                            )
                            stencil_roll.append(0)
                            non_stencil_indices.append(
                                slice(rs.before, None if rs.after == 0 else -rs.after)
                            )
                        case BoundaryCondition.wrap:
                            # a wrapping boundary is compatible with any other
                            #  boundary
                            if r[0] not in (
                                BoundaryCondition.valid,
                                BoundaryCondition.wrap,
                            ):
                                raise (
                                    ValueError(
                                        f"requires {r[0].name} stencil "
                                        + "boundary but found wrapping boundary"
                                    )
                                    | ctx
                                )

                            # what is the required stencil after adjusting for
                            #  near-boundary stencil truncation?
                            # (a) if the chunk is in the middle, where no
                            #     boundary condition is applied, we just keep
                            #     the stencil as-is
                            # (b) if the chunk's stencil only overlaps with the
                            #     boundary on one side, we keep the stencil on
                            #     that side as-is, as long as it does not
                            #     overlap, after wrap-around, with the stencil
                            #     on the other side in this case, we also need
                            #     to roll the stencil on the side of the data
                            #     boundary to the other side, to ensure that
                            #     other boundary conditions also see a boundary
                            # (c) otherwise, we can have the full data and
                            #     remove any excessive stencil so that the
                            #     per-safeguard stencil correctly sees how
                            #     points wrap
                            rs = NeighbourhoodAxis(
                                before=(
                                    r[1].before  # (a)
                                    if r[1].before <= chunk_offset[i]
                                    else (
                                        min(  # (b)
                                            r[1].before - chunk_offset[i],
                                            data_shape[i]
                                            - chunk_offset[i]
                                            - chunk_shape[i]
                                            - r[1].after,
                                        )
                                        if r[1].after
                                        <= (
                                            data_shape[i]
                                            - chunk_shape[i]
                                            - chunk_offset[i]
                                        )
                                        else min(chunk_offset[i], r[1].before)  # (c)
                                    )
                                ),
                                after=(
                                    r[1].after  # (a)
                                    if r[1].after
                                    <= (
                                        data_shape[i] - chunk_shape[i] - chunk_offset[i]
                                    )
                                    else (
                                        min(  # (b)
                                            chunk_offset[i]
                                            + chunk_shape[i]
                                            + r[1].after
                                            - data_shape[i],
                                            chunk_offset[i] - r[1].before,
                                        )
                                        if r[1].before <= chunk_offset[i]
                                        else min(  # (c)
                                            r[1].after,
                                            data_shape[i]
                                            - chunk_shape[i]
                                            - chunk_offset[i],
                                        )
                                    )
                                ),
                            )
                            if c[1].before < rs.before:
                                with ctx.parameter("before"):
                                    raise (
                                        ValueError(
                                            "requires stencil with at least "
                                            + f"{rs.before} point(s) before "
                                            + f"but received only {c[1].before}"
                                        )
                                        | ctx
                                    )
                            if c[1].after < rs.after:
                                with ctx.parameter("after"):
                                    raise (
                                        ValueError(
                                            "requires stencil with at least "
                                            + f"{rs.after} point(s) after but "
                                            + f"received only {c[1].after}"
                                        )
                                        | ctx
                                    )

                            roll_before = max(0, rs.before - chunk_offset[i])
                            roll_after = max(
                                0,
                                chunk_offset[i]
                                + chunk_shape[i]
                                + rs.after
                                - data_shape[i],
                            )
                            if (roll_before > 0) and (roll_after > 0):
                                roll_before, roll_after = 0, 0

                            stencil_indices.append(
                                slice(
                                    c[1].before - rs.before,
                                    None
                                    if c[1].after == rs.after
                                    else rs.after - c[1].after,
                                )
                            )
                            stencil_roll.append(-roll_before + roll_after)
                            nsi_before = rs.before + roll_after - roll_before
                            nsi_after = rs.after + roll_before - roll_after
                            non_stencil_indices.append(
                                slice(
                                    nsi_before, None if nsi_after == 0 else -nsi_after
                                )
                            )
                        case _:
                            assert_never(c[0])

        data_chunk_ = np.roll(
            data_chunk[tuple(stencil_indices)],
            shift=tuple(stencil_roll),
            axis=tuple(range(data_chunk.ndim)),
        )
        prediction_chunk_ = np.roll(
            prediction_chunk[tuple(stencil_indices)],
            shift=tuple(stencil_roll),
            axis=tuple(range(prediction_chunk.ndim)),
        )

        # create the late-bound bindings for the chunk
        late_bound_chunk = (
            late_bound_chunk
            if isinstance(late_bound_chunk, Bindings)
            else Bindings(**late_bound_chunk)
        )
        # check that all late-bound parameters have the right shape
        late_bound_chunk.expect_broadcastable_to(data_chunk.shape)
        # apply the stencil indices to the late-bound parameters
        late_bound_chunk = late_bound_chunk.apply_slice_index(tuple(stencil_indices))
        late_bound_chunk = late_bound_chunk.apply_roll(tuple(stencil_roll))

        late_bound_reqs = self.late_bound
        late_bound_builtin = {
            p: data_chunk_ for p in late_bound_reqs if p in self.builtin_late_bound
        }
        late_bound_reqs = frozenset(late_bound_reqs - late_bound_builtin.keys())
        late_bound_keys = frozenset(late_bound_chunk.parameters())
        LateBoundParameterResolutionError.check_or_raise(
            late_bound_reqs, late_bound_keys
        )

        if len(late_bound_builtin) > 0:
            late_bound_chunk = late_bound_chunk.update(
                **{str(p): v for p, v in late_bound_builtin.items()}
            )

        where_chunk_: Literal[True] | np.ndarray[tuple[int, ...], np.dtype[np.bool]] = (
            True
            if where_chunk is True
            else np.roll(
                where_chunk[tuple(stencil_indices)],
                shift=tuple(stencil_roll),
                axis=tuple(range(where_chunk.ndim)),
            )
        )

        return (
            data_chunk_,
            prediction_chunk_,
            late_bound_chunk,
            where_chunk_,
            tuple(non_stencil_indices),
        )

    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the safeguards.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the safeguards.
        """

        return dict(
            safeguards=[safeguard.get_config() for safeguard in self.safeguards],
            _version=str(self.version),
        )

    @classmethod
    def from_config(cls, config: dict[str, JSON]) -> Self:
        """
        Instantiate the safeguards from a configuration [`dict`][dict].

        Parameters
        ----------
        config : dict[str, JSON]
            Configuration of the safeguards.

        Returns
        -------
        safeguards : Self
            Collection of safeguards.
        """

        return cls(**config)  # type: ignore

    @override
    def __repr__(self) -> str:
        return f"{type(self).__name__}(safeguards={list(self.safeguards)!r})"


_SAFEGUARDS_VERSION: Final[Version] = Version(1, 0, 0)


_SUPPORTED_DTYPES: Final[frozenset[np.dtype[np.number]]] = frozenset(
    {
        np.dtype(np.int8),
        np.dtype(np.int16),
        np.dtype(np.int32),
        np.dtype(np.int64),
        np.dtype(np.uint8),
        np.dtype(np.uint16),
        np.dtype(np.uint32),
        np.dtype(np.uint64),
        np.dtype(np.float16),
        np.dtype(np.float32),
        np.dtype(np.float64),
    }
)
